# Base imports
import numpy as np
import pandas as pd

# sklearn imports
from sklearn.metrics import mean_absolute_error

# utils
from ...utils import mat_to_binary, normalize_tensor

# Recommender Efficacy Metrics
from ...utils._recommender_tools import (
    avg_f1,
    avg_precision,
    avg_recall,
    entropy,
    recommender_mae,
    recommender_rmse,
)
from ...utils._validation import _recommender_checks


def aggregate_diversity(mat_pred, top=None, thresh=0.5, normalize=False):
    r"""
    Aggregate Diversity.

    Description
    ----------
    Given a matrix of scores, this function computes the recommended items for
    each user, selecting either the highest-scored items or those above an input
    threshold. It then returns the aggregate diversity: the proportion of recommended
    items out of all possible items.

    Interpretation
    --------------
    A value of 1 is desired. We wish for a high proportion of items
    to be shown to avoid the 'rich get richer effect'.

    Parameters
    ----------
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    top (optional) : int
        If not None, the number of items recommended to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Aggregate Diversity : :math:`\frac{|Items\; shown|}{|Items|}`

    References
    ----------
    .. [1] `H Abdollahpouri and M Mansoury and R Burke and B Mobasher and E Malthouse (2021).
            User-centered Evaluation of Popularity Bias in Recommender Systems, ACM.
            <https://doi.org/10.1145%2F3450613.3456821>`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import aggregate_diversity
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> aggregate_diversity(mat_pred, top=None, thresh=0.8, normalize=True)
    0.75
    """
    # input checks and coerce
    _, _, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=None,
        group_b=None,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalize scores
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # Make matrix binary
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)

    # Count items by summing over users
    item_count = binary_mat_pred.sum(axis=0)

    # Proportion of all items shown
    agg_div = (item_count >= 1).sum() / len(item_count)

    return agg_div


def gini_index(mat_pred, top=None, thresh=0.5, normalize=False):
    """
    GINI index.

    Description
    ----------
    Measures the inequality across the frequency distribution
    of the recommended items.

    Interpretation
    --------------
    An algorithm that recommends each item the same number of
    times (uniform distribution) will have a Gini index of 0
    and the one with extreme inequality will have a Gini of 1.

    Parameters
    ----------
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        GINI

    References
    ----------
    .. [1] `M Mansoury, H Abdollahpouri, M Pechenizkiy, B Mobasher, R Burke (2020).
            FairMatch: A Graph-based Approach for Improving Aggregate Diversity in Recommender Systems,
            <https://doi.org/10.48550/arXiv.2005.01148>`
    .. [2] `Farzad Eskandanian, Bamshad Mobasher (2020).
            Using Stable Matching to Optimize the Balance between Accuracy and Diversity in Recommendation
            <https://doi.org/10.48550/arXiv.2006.03715>`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import gini_index
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> gini_index(mat_pred, top=2, thresh=None, normalize=False)
    0.1333333333333333
    """
    # input check and coerce
    _, _, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=None,
        group_b=None,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalize score matrix
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # Make matrix binary
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)

    # compute frequencies and sort them
    item_nums = binary_mat_pred.sum(axis=0)
    item_freqs = item_nums / item_nums.sum()
    item_freqs_s = np.sort(item_freqs)

    # compute gini sum
    num_items = mat_pred.shape[1]
    w = 2 * np.arange(num_items) - num_items + 1
    gini_val = (w * item_freqs_s).sum() / (num_items - 1)

    return gini_val


def exposure_entropy(mat_pred, top=None, thresh=0.5, normalize=False):
    r"""
    Exposure Entropy.

    Description
    ----------
    This function measures the entropy of the item exposure distribution.

    Interpretation
    --------------
    A low entropy (close to 0) indicates high certainty as to which item
    will be shown. Higher entropies therefore ensure a more
    homogeneous distribution. Scale is relative to number of items.

    Parameters
    ----------
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Exposure Entropy : :math:`-\sum_{k}{ p_k} \ln(p_k)`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import exposure_entropy
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> exposure_entropy(mat_pred, top=None, thresh=0.3, normalize=True)
    1.3762266043445464
    """
    # input checks and coerce
    _, _, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=None,
        group_b=None,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalize score matrix
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # Get the item exposures
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)
    item_exposures = binary_mat_pred.sum(axis=0)
    item_exposure_dist = item_exposures / item_exposures.sum()

    # Return entropy
    return entropy(item_exposure_dist)


def avg_recommendation_popularity(mat_pred, top=None, thresh=0.5, normalize=False):
    """
    Average Recommendation Popularity.

    Description
    ----------
    This function computes the average recommendation popularity
    of items over users. We define the recommendation popularity
    as the average amount of times an item is recommended.

    Interpretation
    --------------
    A low value is desidered and suggests that items have been
    recommended equally across the population.

    Parameters
    ----------
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Average Recommendation Popularity

    References
    ----------
    .. [1] `M Mansoury, H Abdollahpouri, M Pechenizkiy, B Mobasher, R Burke, E Malthouse (2020).
            User-centered Evaluation of Popularity Bias in Recommender Systems,
            <https://arxiv.org/pdf/2103.06364.pdf>`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import avg_recommendation_popularity
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> avg_recommendation_popularity(mat_pred, top=None, thresh=0.5, normalize=False)
    5.037037037037036
    """
    # input checks and coerce
    _, _, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=None,
        group_b=None,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalize score matrix
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # Make matrices binary
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)
    item_count = binary_mat_pred.sum(axis=0)

    val = (binary_mat_pred * item_count).sum(axis=1) / (binary_mat_pred.sum(axis=1))
    return np.nanmean(val)


def mad_score(group_a, group_b, mat_pred, normalize=False):
    r"""
    Mean Absolute Deviation.

    Description
    ----------
    Difference in average score for group_a and group_b.

    Interpretation
    --------------
    A large value of MAD indicates differential treatment of
    group a and group b. A positive value indicates that
    group a received higher scores on average, while
    a negative value indicates higher ratings for group b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        MAD Score : :math:`\texttt{avg_group_a - avg_group_b}`

    References
    ----------
    .. [1] `Ziwei Zhu, Xia Hu, and James Caverlee (2018).
            Fairness-Aware Tensor-Based Recommendation,
            <https://dl.acm.org/doi/pdf/10.1145/3269206.3271795>`
    .. [2] `Y Deldjoo, V W Anelli, H Zamani, A Bellogın, TDi, T D Noia (2021).
            A Flexible Framework for Evaluating User and Item Fairness in Recommender Systems,
            <https://link.springer.com/article/10.1007/s11257-020-09285-1>`
    .. [3] `Y Deldjooa, A Bellogin, T D Noiaa (2021).
            Explaining recommender systems fairness and accuracy through the lens of data characteristics,
            <https://www.sciencedirect.com/science/article/pii/S0306457321001503>`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import mad_score
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                            [0.7, 0.9, 0.1, 0.7],
                            [0.3, 0.2, 0.3, 0.3],
                            [0.2, 0.1, 0.7, 0.8],
                            [0.8, 0.7, 0.9, 0.1],
                            [1. , 0.9, 0.3, 0.6],
                            [0.8, 0.9, 0.1, 0.1],
                            [0.2, 0.3, 0.1, 0.5],
                            [0.1, 0.2, 0.7, 0.7],
                            [0.2, 0.7, 0.1, 0.2]])
    >>> mad_score(group_a, group_b, mat_pred, normalize=False)
    0.00833333333333336
    """
    # input checks and coerce
    group_a, group_b, mat_pred, _, _, _, normalize = _recommender_checks(
        group_a=group_a,
        group_b=group_b,
        mat_pred=mat_pred,
        mat_true=None,
        top=None,
        thresh=None,
        normalize=normalize,
    )

    # normalize
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # Split by group
    mat_pred_a = mat_pred[group_a == 1]
    mat_pred_b = mat_pred[group_b == 1]

    # Get averages
    avg_a = np.nanmean(mat_pred_a)
    avg_b = np.nanmean(mat_pred_b)

    return avg_a - avg_b


def exposure_l1(group_a, group_b, mat_pred, top=None, thresh=0.5, normalize=False):
    """
    Exposure Total Variation.

    Description
    ----------
    This function computes the total variation norm between the group_a
    exposure distribution to the group_b exposure distribution.

    Interpretation
    --------------
    A total variation divergence of 0 is desired, which occurs when the distributions
    are equal. The maximum value is 1 indicating the distributions are
    very far apart.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Exposure Total Variation

    References
    ----------
    .. [1] `T Giannakas, P Sermpezis, A Giovanidis, T Spyropoulos, G Arvanitakis (2021).
            Fairness in Network-Friendly Recommendations,
            <https://arxiv.org/pdf/2104.00959.pdf>`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import exposure_l1
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                            [0.7, 0.9, 0.1, 0.7],
                            [0.3, 0.2, 0.3, 0.3],
                            [0.2, 0.1, 0.7, 0.8],
                            [0.8, 0.7, 0.9, 0.1],
                            [1. , 0.9, 0.3, 0.6],
                            [0.8, 0.9, 0.1, 0.1],
                            [0.2, 0.3, 0.1, 0.5],
                            [0.1, 0.2, 0.7, 0.7],
                            [0.2, 0.7, 0.1, 0.2]])
    >>> exposure_l1(group_a, group_b, mat_pred, top=1, thresh=None, normalize=False)
    0.25
    """
    # input checks and coerce
    group_a, group_b, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=group_a,
        group_b=group_b,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalize score matrix
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)

    # Split by group
    mat_pred_a = binary_mat_pred[group_a == 1]
    mat_pred_b = binary_mat_pred[group_b == 1]

    # Get the item exposure distribution for group_a
    item_count_a = mat_pred_a.sum(axis=0)
    item_dist_a = item_count_a / item_count_a.sum()

    # Get the item exposure distribution for group_b
    item_count_b = mat_pred_b.sum(axis=0)
    item_dist_b = item_count_b / item_count_b.sum()

    # Compute Total variation distance
    return 0.5 * mean_absolute_error(item_dist_a, item_dist_b) * len(item_dist_a)


def exposure_kl(group_a, group_b, mat_pred, top=None, thresh=0.5, normalize=False):
    """
    Exposure KL Divergence.

    Description
    ----------
    This function computes the KL divergence from the group_a
    exposure distribution to the group_b exposure distribution.

    Interpretation
    --------------
    A KL divergence of 0 is desired, which occurs when the distributions
    are equal. Higher values of the KL divergence indicate difference
    in exposure distributions of group a and group b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        `KL(exp_min,exp_maj)`

    References
    ----------
    .. [1] `A Dash, A Chakraborty, S Ghosh, A Mukherjee, K P. Gummadi (2021).
            When the Umpire is also a Player: Bias in Private Label Product
            Recommendations on E-commerce Marketplaces,
            <https://arxiv.org/pdf/2102.00141.pdf >`
    .. [2] `T Giannakas, P Sermpezis, A Giovanidis, T Spyropoulos, G Arvanitakis (2021).
            Fairness in Network-Friendly Recommendations,
            <https://arxiv.org/pdf/2102.00141.pdf >`


    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import exposure_kl
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                            [0.7, 0.9, 0.1, 0.7],
                            [0.3, 0.2, 0.3, 0.3],
                            [0.2, 0.1, 0.7, 0.8],
                            [0.8, 0.7, 0.9, 0.1],
                            [1. , 0.9, 0.3, 0.6],
                            [0.8, 0.9, 0.1, 0.1],
                            [0.2, 0.3, 0.1, 0.5],
                            [0.1, 0.2, 0.7, 0.7],
                            [0.2, 0.7, 0.1, 0.2]])
    >>> exposure_kl(group_a, group_b, mat_pred, top=1, thresh=None, normalize=False)
    0.23217831296817806
    """
    # input checks and coerce
    group_a, group_b, mat_pred, _, top, thresh, normalize = _recommender_checks(
        group_a=group_a,
        group_b=group_b,
        mat_pred=mat_pred,
        mat_true=None,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # normalize score matrix
    if normalize:
        mat_pred = normalize_tensor(mat_pred)

    # Split by group
    mat_pred_a = mat_pred[group_a == 1]
    mat_pred_b = mat_pred[group_b == 1]

    # Get the item exposure distribution for group_a
    binary_mat_pred_a = mat_to_binary(mat_pred_a, top=top, thresh=thresh)
    item_count_a = binary_mat_pred_a.sum(axis=0)
    item_dist_a = item_count_a / item_count_a.sum()

    # Get the item exposure distribution for group_b
    binary_mat_pred_b = mat_to_binary(mat_pred_b, top=top, thresh=thresh)
    item_count_b = binary_mat_pred_b.sum(axis=0)
    item_dist_b = item_count_b / item_count_b.sum()

    # Compute KL divergence between dists
    return entropy(item_dist_a, item_dist_b)


def _recommender_metric_ratio(
    metric, group_a, group_b, mat_pred, mat_true, top=None, thresh=0.5, normalize=False
):
    """
    Metric ratio for recommender systems

    Description
    ----------
    This function computes the ratio of a given metric on minority and majority group.

    Parameters
    ----------
    metric : function
        Metric to compute
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    mat_true : matrix-like
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each user,item pair.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Ratio of metrics : Metric(min)/Metric(maj)
    """
    # input checks and coerce
    group_a, group_b, mat_pred, mat_true, top, thresh, normalize = _recommender_checks(
        group_a=group_a,
        group_b=group_b,
        mat_pred=mat_pred,
        mat_true=mat_true,
        top=top,
        thresh=thresh,
        normalize=normalize,
    )

    # if normalize
    if normalize:
        tens = np.stack((mat_pred, mat_true))
        norm_tens = normalize_tensor(tens)
        mat_pred, mat_true = norm_tens

    # Split by group
    mat_pred_a = mat_pred[group_a == 1]
    mat_pred_b = mat_pred[group_b == 1]
    mat_true_a = mat_true[group_a == 1]
    mat_true_b = mat_true[group_b == 1]

    # compute metrics
    if top is None and thresh is None:
        metric_a = metric(mat_pred_a, mat_true_a)
        metric_b = metric(mat_pred_b, mat_true_b)
    else:
        # metrics that have top and thresh as input
        metric_a = metric(mat_pred_a, mat_true_a, top, thresh)
        metric_b = metric(mat_pred_b, mat_true_b, top, thresh)

    # ratio
    return metric_a / metric_b


def avg_precision_ratio(
    group_a, group_b, mat_pred, mat_true, top=None, thresh=0.5, normalize=False
):
    r"""
    Average precision ratio

    Description
    ----------
    This function computes the ratio of average precision (over users)
    on minority and majority group.

    Interpretation
    --------------
    A value of 1 is desired. Lower values show bias against minority group.
    Higher values show bias against majority group.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    mat_true : matrix-like
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each user,item pair.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Ratio of average precisions : :math:`\frac{\texttt{AVg_precision_min}}{\texttt{AVg_precision_maj}}`

    References
    ----------
    .. [1] `Yunqi Li, Hanxiong Chen, Zuohui Fu, Yingqiang Ge, Yongfeng Zhang (2021).
            User-oriented Fairness in Recommendation.
            <https://arxiv.org/pdf/2104.10671.pdf>`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import avg_precision_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> mat_true = np.array([[0.7, 0.8, 0.4, 0.2],
                          [0.9, 0.9, 0.1, 0.2],
                          [0.3, 0.8, 0.2, 0.6],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.6, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.1, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.1, 0.1, 0.8]])
    >>> avg_precision_ratio(group_a, group_b, mat_pred, mat_true, top=None, thresh=0.2, normalize=False)
    1.161290322580645
    """
    return _recommender_metric_ratio(
        avg_precision, group_a, group_b, mat_pred, mat_true, top, thresh, normalize
    )


def avg_recall_ratio(
    group_a, group_b, mat_pred, mat_true, top=None, thresh=0.5, normalize=False
):
    r"""
    Average recall ratio

    Description
    ----------
    This function computes the ratio of average recall (over users)
    on minority and majority group.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    mat_true : matrix-like
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each user,item pair.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Ratio of average recalls : :math:`\frac{\texttt{AVg_recall_min}}{\texttt{AVg_recall_maj}}`

    References
    ----------
    .. [1] `Yunqi Li, Hanxiong Chen, Zuohui Fu, Yingqiang Ge, Yongfeng Zhang (2021).
            User-oriented Fairness in Recommendation.
            <https://arxiv.org/pdf/2104.10671.pdf>`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import avg_recall_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> mat_true = np.array([[0.7, 0.8, 0.4, 0.2],
                          [0.9, 0.9, 0.1, 0.2],
                          [0.3, 0.8, 0.2, 0.6],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.6, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.1, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.1, 0.1, 0.8]])
    >>> avg_recall_ratio(group_a, group_b, mat_pred, mat_true, top=2, thresh=0.5, normalize=False)
    1.0
    """
    return _recommender_metric_ratio(
        avg_recall, group_a, group_b, mat_pred, mat_true, top, thresh, normalize
    )


def avg_f1_ratio(
    group_a, group_b, mat_pred, mat_true, top=None, thresh=0.5, normalize=False
):
    r"""
    Average f1 ratio

    Description
    ----------
    This function computes the ratio of average f1 (over users)
    on minority and majority group.

    Interpretation
    --------------
    A value of 1 is desired. Lower values show bias against minority group.
    Higher values show bias against majority group.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    mat_true : matrix-like
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each user,item pair.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Ratio of average f1 : :math:`\frac{\texttt{AVg_f1_min}}{\texttt{AVg_f1_maj}}`

    References
    ----------
    .. [1] `Yunqi Li, Hanxiong Chen, Zuohui Fu, Yingqiang Ge, Yongfeng Zhang (2021).
            User-oriented Fairness in Recommendation.
            <https://arxiv.org/pdf/2104.10671.pdf>`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import avg_f1_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> mat_true = np.array([[0.7, 0.8, 0.4, 0.2],
                          [0.9, 0.9, 0.1, 0.2],
                          [0.3, 0.8, 0.2, 0.6],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.6, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.1, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.1, 0.1, 0.8]])
    >>> avg_f1_ratio(group_a, group_b, mat_pred, mat_true, top=None, thresh=0.5, normalize=False)
    0.9285714285714286
    """

    return _recommender_metric_ratio(
        avg_f1, group_a, group_b, mat_pred, mat_true, top, thresh, normalize
    )


def recommender_rmse_ratio(group_a, group_b, mat_pred, mat_true, normalize=False):
    """
    Recommender RMSE ratio.

    Description
    ----------
    This function computes the ratio of rmse between
    predictions and scores for group_a and group_b.

    Interpretation
    --------------
    A value of 1 is desired. Lower values show bias against group_a.
    Higher values show bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    mat_true : matrix-like
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each user,item pair.
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Recommender RMSE ratio : :math:`\frac{\texttt{AVg_rmse_min}}{\texttt{AVg_rmse_maj}}`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import recommender_rmse_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> mat_true = np.array([[0.7, 0.8, 0.4, 0.2],
                          [0.9, 0.9, 0.1, 0.2],
                          [0.3, 0.8, 0.2, 0.6],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.6, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.1, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.1, 0.1, 0.8]])
    >>> recommender_rmse_ratio(group_a, group_b, mat_pred, mat_true)
    1.149630441384884
    """
    return _recommender_metric_ratio(
        recommender_rmse,
        group_a,
        group_b,
        mat_pred,
        mat_true,
        top=None,
        thresh=None,
        normalize=normalize,
    )


def recommender_mae_ratio(group_a, group_b, mat_pred, mat_true, normalize=False):
    """
    Recommender MAE ratio.

    Description
    ----------
    This function computes the ratio of mae between
    predictions and scores for group_a and group_b.

    Interpretation
    --------------
    A value of 1 is desired. Lower values show bias against group_a.
    Higher values show bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    mat_true : matrix-like
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each user,item pair.
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.

    Returns
    -------
    float
        Recommender MAE ratio : :math:`\frac{\texttt{AVg_mae_min}}{\texttt{AVg_mae_maj}}`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import recommender_mae_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> mat_pred = np.array([[0.9, 0.8, 0.4, 0.2],
                          [0.7, 0.9, 0.1, 0.7],
                          [0.3, 0.2, 0.3, 0.3],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.8, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.9, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.7, 0.1, 0.2]])
    >>> mat_true = np.array([[0.7, 0.8, 0.4, 0.2],
                          [0.9, 0.9, 0.1, 0.2],
                          [0.3, 0.8, 0.2, 0.6],
                          [0.2, 0.1, 0.7, 0.8],
                          [0.6, 0.7, 0.9, 0.1],
                          [1. , 0.9, 0.3, 0.6],
                          [0.8, 0.1, 0.1, 0.1],
                          [0.2, 0.3, 0.1, 0.5],
                          [0.1, 0.2, 0.7, 0.7],
                          [0.2, 0.1, 0.1, 0.8]])
    >>> recommender_mae_ratio(group_a, group_b, mat_pred, mat_true)
    1.2954545454545452
    """
    return _recommender_metric_ratio(
        recommender_mae,
        group_a,
        group_b,
        mat_pred,
        mat_true,
        top=None,
        thresh=None,
        normalize=normalize,
    )


def recommender_bias_metrics(
    group_a=None,
    group_b=None,
    mat_pred=None,
    mat_true=None,
    top=None,
    thresh=0.5,
    normalize=False,
    metric_type="equal_outcome",
):
    """
    Recommender bias metrics batch computation.

    Description
    ----------
    This function computes all the relevant recommender bias metrics,
    and displays them as a pandas dataframe.

    Parameters
    ----------
    group_a : array-like
        Group membership vector.
    group_b : array-like
        Group membership vector.
    mat_pred : matrix-like
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each user,item interaction.
    mat_true : matrix-like
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each user,item pair.
    top (optional) : int
        If not None, the number of items that are shown to each user.
    thresh (optional) : float
        Threshold indicating value at which
        a given item is shown to user (if top is None).
    normalize (optional) : bool
        If True, normalises the data matrix to [0,1] range.
    metric_type : 'all', 'item_based', 'equal_outcome' or 'equal_opportunity'
        Specifies which metrics we compute

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """
    item_perform = {
        "Aggregate Diversity": aggregate_diversity,
        "GINI index": gini_index,
        "Exposure Distribution Entropy": exposure_entropy,
        "Average Recommendation Popularity": avg_recommendation_popularity,
    }

    group_perform = {
        "Mean Absolute Deviation": mad_score,
    }

    group2_perform = {
        "Exposure Total Variation": exposure_l1,
        "Exposure KL Divergence": exposure_kl,
    }

    group_true_perform = {
        "Average Precision Ratio": avg_precision_ratio,
        "Average Recall Ratio": avg_recall_ratio,
        "Average F1 Ratio": avg_f1_ratio,
    }

    group_true_reg_perform = {
        "Recommender RMSE Ratio": recommender_rmse_ratio,
        "Recommender MAE Ratio": recommender_mae_ratio,
    }

    ref_vals = {
        "Aggregate Diversity": 1,
        "GINI index": 0,
        "Exposure Distribution Entropy": "-",
        "Average Recommendation Popularity": "-",
        "Mean Absolute Deviation": 0,
        "Exposure Total Variation": 0,
        "Exposure KL Divergence": 0,
        "Average Precision Ratio": 1,
        "Average Recall Ratio": 1,
        "Average F1 Ratio": 1,
        "Recommender RMSE Ratio": 1,
        "Recommender MAE Ratio": 1,
    }

    item_metrics = []
    out_metrics = []
    opp_metrics = []

    item_metrics += [
        [pf, fn(mat_pred, top, thresh, normalize), ref_vals[pf]]
        for pf, fn in item_perform.items()
    ]
    if group_a is not None:
        out_metrics += [
            [pf, fn(group_a, group_b, mat_pred, normalize), ref_vals[pf]]
            for pf, fn in group_perform.items()
        ]
        out_metrics += [
            [pf, fn(group_a, group_b, mat_pred, top, thresh, normalize), ref_vals[pf]]
            for pf, fn in group2_perform.items()
        ]
    if mat_true is not None:
        opp_metrics += [
            [
                pf,
                fn(group_a, group_b, mat_pred, mat_true, top, thresh, normalize),
                ref_vals[pf],
            ]
            for pf, fn in group_true_perform.items()
        ]
        opp_metrics += [
            [pf, fn(group_a, group_b, mat_pred, mat_true, normalize), ref_vals[pf]]
            for pf, fn in group_true_reg_perform.items()
        ]

    if metric_type == "all":
        metrics = item_metrics + out_metrics + opp_metrics
        return pd.DataFrame(
            metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    if metric_type == "item_based":
        return pd.DataFrame(
            item_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    if metric_type == "equal_outcome":
        return pd.DataFrame(
            out_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    if metric_type == "equal_opportunity":
        return pd.DataFrame(
            opp_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    else:
        raise ValueError(
            "metric_type is not one of : all, item_based, equal_outcome, equal_opportunity"
        )



import math

import numpy as np
from tqdm import tqdm


class FairRecAlg:
    def __init__(self, rec_size=10, MMS_fraction=0.5, verbose=0):
        self.rec_size = rec_size
        self.MMS_fraction = MMS_fraction
        self.verbose = verbose

    def rank(self, X):
        self.m = X.shape[0]
        self.n = X.shape[1]
        U = list(range(self.m))
        P = list(range(self.n))
        A = {u: [] for u in U}
        F = {u: P[:] for u in U}
        l = int(self.MMS_fraction * self.m * self.rec_size / (self.n + 0.0))
        R = int(math.ceil((l * self.n) / (self.m + 0.0)))
        T = l * self.n
        [B, F1] = self._greedy_round_robin(R, l, T, X, U[:], F.copy())

        F = F1.copy()

        for u in U:
            A[u] = A[u][:] + B[u][:]
        u_less = []
        for u in A:
            if len(A[u]) < self.rec_size:
                u_less.append(u)
        for u in u_less:
            scores = X[u, :]
            new = scores.argsort()[-(self.rec_size + self.rec_size) :][::-1]
            for p in new:
                if p not in A[u]:
                    A[u].append(p)
                if len(A[u]) == self.rec_size:
                    break

        return A

    def _greedy_round_robin(self, R, l, T, V, U, F):
        """
        Description
        -----------
        Greedy round robin allocation based on a specific ordering of
        customers (assuming the ordering is done in the relevance scoring
        matrix before passing it here)
        """

        B = {u: [] for u in U}
        Z = {}
        P = list(range(self.n))
        for p in P:
            Z[p] = l
        for t in range(1, R + 1):
            for i in range(self.m):
                if T == 0:
                    return B, F
                u = U[i]
                possible = [(Z[p] > 0) * (p in F[u]) * V[u, p] for p in range(self.n)]
                p_ = np.argmax(possible)
                if (Z[p_] > 0) and (p_ in F[u]) and len(F[u]) > 0:
                    F[u] = list(F[u])
                    B[u].append(p_)
                    F[u].remove(p_)
                    Z[p_] = Z[p_] - 1
                    T = T - 1
                else:
                    return B, F
        return B, F



# Base Imports
import warnings

import numpy as np
import pandas as pd


def _check_same_shape(list_of_arr, names=""):
    """
    Check same shape

    Description
    ----------
    This function checks if all numpy arrays
    in a list have same length

    Parameters
    ----------
    list_of_arr : list of numpy arrays
        input
    names : str
        the name of the inputs

    Returns
    -------
    ValueError or None
    """
    num_dims = len(list_of_arr[0].shape)
    for i in range(num_dims):
        try:
            n = len(np.unique([x.shape[i] for x in list_of_arr]))
            if n > 1:
                raise ValueError(names + " do not all have the same shape.")
        except:
            raise ValueError(names + " do not all have the same shape.")


def _check_non_empty(arr, name="", quantile=0):
    """
    Check non empty

    Description
    ----------
    This function checks if a numpy array has
    a sum of 0.

    Parameters
    ----------
    arr : binary array
        input
    name : str
        the name of the input
    quantile : float
        quantile

    Returns
    -------
    warning or None
    """
    num_a = arr.sum()
    if num_a == 0:
        warnings.warn(name + " has no members at quantile " + str(quantile))


def _check_classes_input(classes, y_pred, y_true=None):
    """
    Check classes input (multiclass)

    Description
    ----------
    This function checks the input classes is
    composed of the unique values of y_pred (and y_true)

    Parameters
    ----------
    classes : numpy array
        Class vector (categorical)
    y_pred : numpy array
        Predictions vector (categorical)
    y_true : numpy array
        Target vector (categorical)

    Returns
    -------
    ValueError or None
    """
    # case 1 : y_true is None
    if y_true is None:
        _classes = set(classes)
        __classes = set(y_pred)
        if _classes != __classes:
            raise ValueError("classes is not a reordering of unique values in y_pred.")
    # case 2
    else:
        _classes = set(classes)
        __classes = set(y_pred).union(set(y_true))
        if _classes != __classes:
            raise ValueError(
                "classes is not a reordering of unique values in y_pred or y_true."
            )


def _check_groups_input(groups, p_attr):
    """
    Check groups input (multiclass)

    Description
    ----------
    This function checks the groups input is
    composed of the unique values of p_attr

    Parameters
    ----------
    groups : numpy array
        Class vector (categorical)
    p_attr : numpy array
        Predictions vector (categorical)

    Returns
    -------
    ValueError or None
    """
    # compare sets
    _groups = set(groups)
    __groups = set(p_attr)
    if _groups != __groups:
        raise ValueError("groups is not a reordering of unique values in p_attr.")


def _check_binary(arr, name=""):
    """
    Check binary

    Description
    ----------
    This function checks if a numpy array
    is binary.

    Parameters
    ----------
    arr : numpy array
        input
    name : str
        the name of the input

    Returns
    -------
    ValueError or None
    """
    if not np.array_equal(arr, arr.astype(bool)):
        raise ValueError(name + " is not a binary vector.")


def _check_nan(arr, name=""):
    """
    Check for nan

    Description
    ----------
    This function checks if a numpy array
    has a nan.

    Parameters
    ----------
    arr : numpy array
        input
    name : str
        the name of the input

    Returns
    -------
    ValueError or None
    """
    if pd.isnull(pd.Series(arr)).any():
        raise ValueError(name + " has NaN values.")


def _check_nan_mat(mat, name=""):
    """
    Check for nan

    Description
    ----------
    This function checks if a numpy ndarray
    has a nan.

    Parameters
    ----------
    mat : numpy ndarray
        input
    name : str
        the name of the input

    Returns
    -------
    ValueError or None
    """
    if np.isnan(mat).any():
        return ValueError(name + " has NaN values.")


def _array_like_to_numpy(arr, name=""):
    """
    Coerce input to numpy (if possible)

    Description
    ----------
    This function coerces to numpy where
    possible, and return an error if not.

    Parameters
    ----------
    arr : array-like
        Input to coerce

    Returns
    -------
    numpy array or TypeError
    """
    try:
        out = np.squeeze(np.asarray(arr))
        if len(out.shape) == 1:
            return out
        else:
            raise ValueError()
    except:
        raise TypeError(
            "input {} is not array-like. \
This includes numpy 1d arrays, lists, \
pandas Series or pandas 1d DataFrame".format(
                name
            )
        )


def _matrix_like_to_numpy(arr, name=""):
    """
    Coerce input to numpy (if possible)

    Description
    ----------
    This function coerces to numpy where
    possible, and return an error if not.

    Parameters
    ----------
    arr : matrix-like
        Input to coerce

    Returns
    -------
    numpy array or TypeError
    """
    try:
        out = np.squeeze(np.asarray(arr))
        if len(out.shape) == 2:
            return out
        else:
            raise ValueError()
    except:
        raise TypeError(
            "input {} is not matrix-like. \
This includes numpy 2d arrays, list of lists \
or pandas 2d DataFrame".format(
                name
            )
        )


def _coerce_and_check_arr(arr, name="input"):
    """
    Coerce and check array-like

    Description
    ----------
    This function coerces to numpy where
    possible, and return an error if not.
    Also checks for nan values.

    Parameters
    ----------
    arr : array-like
        Input to coerce
    name : str
        The name of array

    Returns
    -------
    numpy array or TypeError
    """
    # coerce to numpy if possible
    np_arr = _array_like_to_numpy(arr, name=name)
    # check for nan values
    _check_nan(np_arr, name=name)
    # return
    return np_arr


def _coerce_and_check_mat(mat, name="input"):
    """
    Coerce and check matrix-like

    Description
    ----------
    This function coerces to numpy where
    possible, and return an error if not.
    Also checks for nan values.

    Parameters
    ----------
    mat : matrix-like
        Input to coerce
    name : str
        The name of array

    Returns
    -------
    numpy ndarray or TypeError
    """
    # coerce to numpy if possible
    np_mat = _matrix_like_to_numpy(mat, name=name)
    # check for nan values
    _check_nan_mat(np_mat, name=name)
    # return
    return np_mat


def _coerce_and_check_binary_group(arr, name=""):
    """
    Coerce and check binary vector

    Description
    ----------
    This function coerces to numpy where
    possible, and return an error if not.
    Also checks vector is binary.

    Parameters
    ----------
    arr : array-like
        Input to coerce
    name : str
        The name of array

    Returns
    -------
    numpy ndarray or TypeError
    """
    # coerce to numpy if possible
    np_arr = _array_like_to_numpy(arr, name=name)
    # check binary (also checks for nans)
    _check_binary(np_arr, name=name)
    # return
    return np_arr


def _classification_checks(group_a, group_b, y_pred, y_true=None):
    """
    Classification checks

    Description
    ----------
    This function checks inputs to
    a classification task

    Returns
    -------
    coerced inputs
    """
    group_a = _coerce_and_check_binary_group(group_a, name="group_a")
    group_b = _coerce_and_check_binary_group(group_b, name="group_b")
    y_pred = _coerce_and_check_binary_group(y_pred, name="y_pred")

    if y_true is None:
        _check_same_shape([group_a, group_b, y_pred], names="group_a, group_b, y_pred")
    else:
        y_true = _coerce_and_check_binary_group(y_true, name="y_true")
        _check_same_shape(
            [group_a, group_b, y_pred, y_true], names="group_a, group_b, y_pred, y_true"
        )

    return group_a, group_b, y_pred, y_true


def _regression_checks(group_a, group_b, y_pred, y_true=None, q=None):
    """
    Regression checks

    Description
    ----------
    This function checks inputs to
    a regression task

    Returns
    -------
    coerced inputs
    """
    group_a = _coerce_and_check_binary_group(group_a, name="group_a")
    group_b = _coerce_and_check_binary_group(group_b, name="group_b")
    y_pred = _coerce_and_check_arr(y_pred, name="y_pred")

    if y_true is None:
        _check_same_shape([group_a, group_b, y_pred], names="group_a, group_b, y_pred")
    else:
        y_true = _coerce_and_check_arr(y_true, name="y_true")
        _check_same_shape(
            [group_a, group_b, y_pred, y_true], names="group_a, group_b, y_pred, y_true"
        )

    if q is not None:
        q = np.atleast_1d(np.array(q).squeeze())
        if len(q.shape) > 1:
            raise ValueError("q should be float or array-like")

    return group_a, group_b, y_pred, y_true, q


def _clustering_checks(
    group_a, group_b, y_pred, y_true=None, data=None, centroids=None
):
    """
    Clustering checks

    Description
    ----------
    This function checks inputs to
    a clustering task

    Returns
    -------
    coerced inputs
    """
    group_a = _coerce_and_check_binary_group(group_a, name="group_a")
    group_b = _coerce_and_check_binary_group(group_b, name="group_b")

    if y_pred is not None:
        y_pred = _coerce_and_check_arr(y_pred, name="y_pred")
        _check_same_shape([group_a, group_b, y_pred], names="group_a, group_b, y_pred")

    if y_true is not None:
        y_true = _coerce_and_check_arr(y_true, name="y_true")
        _check_same_shape([y_pred, y_true], names="y_pred, y_true")

    if data is not None:
        data = _coerce_and_check_mat(data, name="data")

    if centroids is not None:
        centroids = _coerce_and_check_mat(centroids, name="centroids")

    if centroids is not None and data is not None:
        if data.shape[1] != centroids.shape[1]:
            raise ValueError(
                "data and centroids datapoints do not have same dimensionality"
            )

    return group_a, group_b, y_pred, y_true, data, centroids


def _recommender_checks(
    group_a=None,
    group_b=None,
    mat_pred=None,
    mat_true=None,
    top=None,
    thresh=None,
    normalize=None,
):
    """
    Recommender checks

    Description
    ----------
    This function checks inputs to
    a recommender task

    Returns
    -------
    coerced inputs
    """
    if group_a is not None:
        group_a = _coerce_and_check_binary_group(group_a, name="group_a")

    if group_b is not None:
        group_b = _coerce_and_check_binary_group(group_b, name="group_b")
        _check_same_shape([group_a, group_b], names="group_a, group_b")

    if mat_pred is not None:
        mat_pred = _matrix_like_to_numpy(mat_pred, name="mat_pred")

    if mat_true is not None:
        mat_true = _matrix_like_to_numpy(mat_true, name="mat_true")
        _check_same_shape([mat_pred, mat_true], names="mat_pred, mat_true")

    if top is not None:
        if not isinstance(top, int):
            raise ValueError("top has to be an integer")

    if normalize is not None:
        normalize = bool(normalize)

    return group_a, group_b, mat_pred, mat_true, top, thresh, normalize


def _multiclass_checks(
    p_attr=None, y_pred=None, y_true=None, groups=None, classes=None
):
    """
    Multiclass checks

    Description
    ----------
    This function checks inputs to
    a multiclass task

    Returns
    -------
    coerced inputs
    """
    if p_attr is not None:
        p_attr = _coerce_and_check_arr(p_attr, name="p_attr")

    if y_pred is not None:
        y_pred = _coerce_and_check_arr(y_pred, name="y_pred")

    if y_true is not None:
        y_true = _coerce_and_check_arr(y_true, name="y_true")

    # length check
    if p_attr is not None and y_pred is not None:
        _check_same_shape([p_attr, y_pred], names="p_attr, y_pred")

    # length check
    if y_pred is not None and y_true is not None:
        _check_same_shape([y_pred, y_true], names="y_pred, y_true")

    # define groups if not defined (where possible)
    if groups is None:
        if p_attr is not None:
            groups = np.sort(np.unique(p_attr))
        else:
            groups = None
    else:
        _check_groups_input(groups, p_attr)

    # define classes if not defined (where possible)
    if classes is None:
        if y_true is not None and y_pred is not None:
            classes = np.sort(np.unique(np.concatenate((y_pred, y_true))))
        elif y_true is None and y_pred is not None:
            classes = np.sort(np.unique(y_pred))
        else:
            classes = None
    else:
        if y_true is not None and y_pred is not None:
            _check_classes_input(classes, y_pred, y_true)
        elif y_true is None and y_pred is not None:
            _check_classes_input(classes, y_pred)
        else:
            classes = None

    return p_attr, y_pred, y_true, groups, classes


def _check_valid_y_proba(y_proba: np.ndarray):
    atol = 1e-3
    y_shape = np.shape(y_proba)
    assert len(y_shape) == 2, f"""y_proba must be 2d tensor, found: {y_shape}"""

    sum_all_probs = np.sum(y_proba, axis=1)
    assert np.all(
        np.isclose(sum_all_probs, 1, atol=atol)
    ), f"""probability must sum to 1 across the possible classes, found: {sum_all_probs}"""

    correct_proba_values = np.all(y_proba <= 1) and np.all(y_proba >= 0)
    assert (
        correct_proba_values
    ), f"""probability values must be in the interval [0,1], found: {y_proba}"""



from holisticai.bias.mitigation.inprocessing.commons import Logging


class ObjectiveFunction:
    """objective function of logistic regression with prejudice remover
    Loss Function type 4: Weights for logistic regression are prepared for each
    value of S. Penalty for enhancing is defined as mutual information between
    Y and S.
    """

    def __init__(self, estimator, loss_fn):
        self.loss_fn = loss_fn
        self.estimator = estimator

    def loss(self, coef_, X, y, groups):
        """loss function: negative log - likelihood with l2 regularizer
        To suppress the warnings at np.log, do "np.seterr(all='ignore')"
        Parameters
        ----------
        `coef_` : array, shape=(`nb_group_values` * n_features)
            coefficients of model
        X : array, shape=(n_samples, n_features)
            feature vectors of samples
        y : array, shape=(n_samples)
            target class of samples
        sensitive_features : array, shape=(n_samples)
            values of sensitive features
        Returns
        -------
        loss : float
            loss function value
        """
        coef = coef_.reshape(self.estimator.nb_group_values, self.estimator.nb_features)
        X = self.estimator._preprocessing_data(X)
        sigma = self.estimator.sigmoid(X=X, groups=groups, coef=coef)
        loss = self.loss_fn(y=y, sigma=sigma, groups=groups, coef=coef)
        return loss

    def grad_loss(self, coef_, X, y, groups):
        """first derivative of loss function
        Parameters
        ----------
        `coef_` : array, shape=(`nb_group_values` * n_features)
            coefficients of model
        X : array, shape=(n_samples, n_features)
            feature vectors of samples
        y : array, shape=(n_samples)
            target class of samples
        s : array, shape=(n_samples)
            values of sensitive features
        Returns
        grad_loss : float
            first derivative of loss function
        """

        coef = coef_.reshape(self.estimator.nb_group_values, self.estimator.nb_features)
        X = self.estimator._preprocessing_data(X)
        sigma = self.estimator.sigmoid(X=X, groups=groups, coef=coef)
        return self.loss_fn.gradient(X=X, y=y, sigma=sigma, groups=groups, coef=coef)


class PRLogger:
    def __init__(self, total_iterations, verbose, print_interval):
        self.step = 0
        self.verbose = verbose
        self.print_interval = print_interval
        self.total_iterations = total_iterations
        self.log_params = [
            ("iteration", int),
        ]
        self.fn_params = {}
        self.logger = Logging(
            log_params=self.log_params,
            total_iterations=self.total_iterations,
            logger_format="iteration",
        )

    def set_log_fn(self, type, **kargs):
        for param_name, param_fn in kargs.items():
            self.log_params.append((param_name, type))
            self.fn_params[param_name] = param_fn

    def callback(self, x):
        self.step += 1
        if self.verbose > 0:
            if (self.step % self.print_interval) == 0 or (
                self.step % self.total_iterations
            ) == 0:
                args = [self.fn_params[pn](x) for pn, _ in self.log_params[1:]]
                self.logger.update(self.step, *args)

    def info(self, message):
        if self.verbose > 0:
            self.logger.info(message)



from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.metrics.pairwise import pairwise_distances, pairwise_distances_argmin

from holisticai.utils.transformers.bias import BMInprocessing as BMImp
from holisticai.utils.transformers.bias import SensitiveGroups

from .algorithms import (
    fair_k_center_APPROX,
    heuristic_greedy_on_each_group,
    heuristic_greedy_till_constraint_is_satisfied,
)

STRATEGIES_CATALOG = {
    "Fair K-Center": fair_k_center_APPROX,
    "Heuristic Greedy by Group": heuristic_greedy_on_each_group,
    "Heuristic Greedy by Constraint": heuristic_greedy_till_constraint_is_satisfied,
}


class FairKCenterClustering(BaseEstimator, BMImp):
    """
    Fair K-Center Clustering inprocessing bias mitigation implements an approximation algorithm
    for the k-centers problem under the fairness contraint with running time linear in the
    size of the dataset and k (number of cluster).

    Reference
    ---------
        Kleindessner, Matthäus, Pranjal Awasthi, and Jamie Morgenstern. "Fair k-center clustering
        for data summarization." International Conference on Machine Learning. PMLR, 2019.
    """

    def __init__(
        self,
        req_nr_per_group: Optional[list] = [200, 200],
        nr_initially_given: Optional[int] = 100,
        strategy: Optional[str] = "Fair K-Center",
        seed: Optional[int] = None,
    ):
        """
        Parameters
        ----------
        req_nr_per_group : list
            Number of cluster for each group that will be founded.
            - Integer-vector of length m with entries in 0,...,k.
            - Sum of all entries must be equal to k (total number of clusters).

        nr_initially_given: int
            Number of initial random centers.

        strategy: Strategy used to compute the cluster centers. Available are:
            - 'Fair K-Center' (default)
            - 'Heuristic Greedy by Group'
            - 'Heuristic Greedy by Constraint'

        seed: int,
            Initial random seed.
        """
        self.req_nr_per_group = np.array(req_nr_per_group)
        self.nr_initially_given = nr_initially_given
        self.strategy = strategy
        self.seed = seed
        self.sensgroup = SensitiveGroups()

    def fit(self, X, group_a, group_b):
        """
        Fit model using Fair K-Center Clustering.

        Parameters
        ----------
        X : matrix-like
            Input matrix
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
            Self
        """
        np.random.seed(self.seed)

        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        sensitive_groups = np.c_[group_a, group_b]
        p_attr = np.array(
            self.sensgroup.fit_transform(sensitive_groups, convert_numeric=True)
        )

        n = len(X)
        dmat = pairwise_distances(X, metric="l1")
        initially_given = np.random.choice(
            n, size=self.nr_initially_given, replace=False
        )
        centers = STRATEGIES_CATALOG[self.strategy](
            dmat, p_attr, self.req_nr_per_group, initially_given
        )
        cost = np.amax(
            np.amin(
                dmat[np.ix_(np.hstack((centers, initially_given)), np.arange(n))],
                axis=0,
            )
        )
        self.centers = centers
        self.initially_given = initially_given
        self.cost = cost

        self.centroids = X[self.centers]
        self.all_centroids = X[self.all_centers]

        self.labels_ = self.all_centers[
            pairwise_distances_argmin(X, Y=self.all_centroids, metric="l1")
        ]
        self.center_groups_ = p_attr[self.all_centers]

    @property
    def all_centers(self):
        return np.concatenate([self.centers, self.initially_given], axis=0)

    def predict(self, X):
        return pairwise_distances(self.all_centroids, X, metric="l1").argmin(axis=0)



import math

import numpy as np

from holisticai.bias.mitigation.inprocessing.commons import Logging


class DebiaserOptimizer:
    def __init__(self, eps0, rho, gamma, sgd_steps, batch_size, y, lambda_0=0, mu_0=0):
        self.eps0 = eps0
        self.rho = rho
        self.sgd_steps = sgd_steps
        self.lambda_k = lambda_0
        self.mu_k = mu_0
        self.batch_size = batch_size
        self.gamma = gamma
        self.y = y
        self.set_alpha()
        self.generator = self._build_generator(y)

    def get_batch(self):
        return next(self.generator)

    def compute_gradients(self, y_batch):
        lambda_minus_mu = self.lambda_k - self.mu_k
        xi_arg = np.maximum(y_batch, lambda_minus_mu)
        xi_arg = np.minimum(xi_arg, self.gamma + lambda_minus_mu)
        mean_xi = (np.mean(xi_arg) - lambda_minus_mu) / self.gamma

        lambda_gradient = self.eps0 + self.rho - mean_xi
        mu_gradient = self.eps0 - self.rho + mean_xi
        gradients = {"lambda": lambda_gradient, "mu": mu_gradient}
        return gradients

    def update_parameters(self, gradients):
        """
        If self.eps=0, we can drop mu_k and optimize lambda_k only but
        lambda_k will not be constrained to be non-negative in this case.
        """
        if self.eps0 > 1e-3:
            self.lambda_k = max(0, self.lambda_k - self.alpha * gradients["lambda"])
            self.mu_k = max(0, self.mu_k - self.alpha * gradients["mu"])
        else:
            self.lambda_k = self.lambda_k - self.alpha * gradients["lambda"]

    def update_step(self):
        y_batch = self.get_batch()
        grads = self.compute_gradients(y_batch)
        self.update_parameters(grads)


class SGDDebiaser(DebiaserOptimizer):
    def set_alpha(self):
        num_samples_sgd = self.sgd_steps * self.batch_size
        lr = self.gamma * math.sqrt(1.0 / num_samples_sgd)
        self.alpha = lr * self.batch_size

    def _build_generator(self, y):
        idx = np.arange(len(y))  # instance IDs in group k
        group_size = len(idx)
        while True:
            batch_ids = np.random.randint(0, group_size, self.batch_size)
            batch_ids = idx[batch_ids]
            yield y[batch_ids]


class FullGDDebiaser(DebiaserOptimizer):
    def set_alpha(self):
        self.alpha = 0.5

    def _build_generator(self, y):
        while True:
            yield y


class RTLogger:
    def __init__(self, total_iterations, verbose):
        self.verbose = verbose
        log_params = [("iteration", int)]
        self.logger = Logging(
            log_params=log_params,
            total_iterations=total_iterations,
            logger_format="iteration",
        )

    def update(self, iteration):
        if self.verbose:
            self.logger.update(iteration)



import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, balanced_accuracy_score


def get_indexes_from_names(df, names):
    """
    Description
    -----------
    Gets the indexes from the columns

    Parameters
    ----------
    df : pandas dataframe
        the complete dataframe

    Returns
    -------
    list : list of index
    """
    indexes = []
    for item in names:
        indexes.append(df.columns.get_loc(item))
    return indexes


def process_y(y):
    """
    Description
    -----------
    Converts from one-hot encoding to a list of labels

    Parameters
    ----------
    y : numpy array
        the labels matrix

    Returns
    -------
    y_processed (list of labels)
    """
    y_processed = []
    for line in y:
        for i in range(len(line)):
            if line[i] == 1:
                y_processed.append(i)
    return y_processed


def get_majority_class(y):
    """
    Description
    -----------
    Converts from one-hot encoding to a list of labels

    Parameters
    ----------
    y : list
        list of labels

    Returns
    -------
    The majority class and its accuracy
    """
    y = np.array(y).flatten()
    most_common_item = max(y, key=y.tolist().count)
    acc = np.count_nonzero(y == most_common_item) / len(y)
    return most_common_item, acc


def get_initial_solution(y):
    """
    Description
    -----------
    Calculates the accuracy of the majority class

    Parameters
    ----------
    y : numpy array
        the labels matrix

    Returns
    -------
    The majority class and its accuracy
    """
    y_processed = process_y(y)
    return get_majority_class(y_processed)


def remove_inconcsistency(x, y):
    """
    Description
    -----------
    Remove the inconsistencies

    Parameters
    ----------
    x : numpy array
        Dataset features
    y : numpy array
        Dataset labels

    Returns
    -------
    The dataset withtout the inconsistencies
    """
    x = x.tolist()
    y = y.tolist()

    new_x = []
    new_y = []

    for i in range(len(x)):
        target = get_max_y(x[i], x, y)
        if y[i][target] == 1:
            new_x.append(x[i])
            new_y.append(y[i])

    return np.array(new_x), np.array(new_y)


def get_max_y(cur_x, x, y):
    y = [y[i] for i, x in enumerate(x) if x == cur_x]

    counts = [0 for i in range(len(y[0]))]

    for i in range(len(y)):
        y[i] = y[i].index(1)

    for i in range(len(y)):
        counts[y[i]] += 1

    return counts.index(max(counts))


def get_class_count(y):
    """
    Description
    -----------
    Get the count for each class in the labels

    Parameters
    ----------
    y : numpy array
        Dataset labels

    Returns
    -------
    The count of each class
    """
    count = []

    for i in range(len(y[0])):
        count.append(0)

    for labels in y:
        for index, label in enumerate(labels):
            if label == 1:
                count[index] += 1

    return count


def get_class_indexes(y):
    """
    Description
    -----------
    Get the indexes of each class in the labels

    Parameters
    ----------
    y : numpy array
        Dataset labels

    Returns
    -------
    The indexes of each class
    """
    indexes = []

    for i in range(len(y[0])):
        indexes.append([])

    for i, labels in enumerate(y):
        for index, label in enumerate(labels):
            if label == 1:
                indexes[index].append(i)

    return indexes


def predict(x, l_lists):
    """
    Description
    -----------
    Get the indexes of each class in the labels

    Parameters
    ----------
    x : numpy array
        Dataset features

    l_lists : list
        set of scoring systems

    Returns
    -------
    The predictions of the set of scoring systems for the given entries
    """
    y = []

    for index, sample in enumerate(x):
        scores = []
        for l_list in l_lists:
            scores.append(sum(feature * l_list[j] for j, feature in enumerate(sample)))
        y_pred = []
        for i in range(len(scores)):
            if i == np.argmax(scores):
                y_pred.append(1)
            else:
                y_pred.append(0)
        y.append(y_pred)

    return np.array(y)


def format_labels(y):
    y_formatted = []

    for labels in y:
        for i in range(len(labels)):
            if labels[i] == 1:
                y_formatted.append(i)

    return y_formatted


def get_accuracy(x, y, l_lists):
    """
    Description
    -----------
    Compute accuracy for the scoring systems

    Parameters
    ----------
    x : numpy array
        Dataset features

    y : numpy array
        Dataset labels

    l_lists : list
        set of scoring systems

    Returns
    -------
    The accuracy of the predictions
    """
    y_pred = predict(x, l_lists)
    y = format_labels(y)
    y_pred = format_labels(y_pred)
    accuracy = accuracy_score(y, y_pred)

    return accuracy


def get_balanced_accuracy(x, y, l_lists):
    y_pred = predict(x, l_lists)
    y = format_labels(y)
    y_pred = format_labels(y_pred)
    balanced_accuracy = balanced_accuracy_score(y, y_pred)

    return balanced_accuracy



import random


class Utils:
    def __init__(self, upper_bound, lower_bound, Nx, k):
        self.upper_bound = upper_bound
        self.lower_bound = lower_bound
        self.Nx = Nx
        self.k = k

    def select_items(self, group_cluster, Cj, mode="upper"):
        T = []
        for js in group_cluster:
            beta = Cj[js]["beta"]
            if mode == "upper":
                threshold = self.upper_bound
            elif mode == "exact":
                threshold = self.Nx // self.k
            else:
                threshold = self.lower_bound

            n_items_to_container = beta - threshold

            index = Cj[js]["index"]
            random.shuffle(index)
            index_selected, index = (
                index[:n_items_to_container],
                index[n_items_to_container:],
            )

            assert len(index) == threshold

            Cj[js]["index"] = index
            Cj[js]["beta"] = len(index)
            T += index_selected
        return T, Cj

    def choose_items(self, group_cluster, Cj, T, mode="upper"):
        for js in group_cluster:
            beta = Cj[js]["beta"]
            if mode == "upper":
                threshold = self.upper_bound
            elif mode == "exact":
                threshold = self.Nx // self.k
            else:
                threshold = self.lower_bound

            n_items_from_container = threshold - beta
            assert n_items_from_container >= 0
            selected_index, T = T[:n_items_from_container], T[n_items_from_container:]
            Cj[js]["index"] += selected_index
            Cj[js]["beta"] = len(Cj[js]["index"])
        return T, Cj



# Imports
from .correlation_remover import CorrelationRemover
from .fairlet_clustering.transformer import FairletClusteringPreprocessing
from .learning_fair_representation import LearningFairRepresentation
from .reweighing import Reweighing

__all__ = [
    "LearningFairRepresentation",
    "Reweighing",
    "CorrelationRemover",
    "FairletClusteringPreprocessing",
]

import importlib

networkx_spec = importlib.util.find_spec("networkx")
if networkx_spec is not None:
    from .disparate_impact_remover import DisparateImpactRemover

__all__ += ["DisparateImpactRemover"]



class WEstimator:
    """
    This wrap the estimator class and helps link parameters with updates done by unconventional transformers during pipeline execution.
    """

    def __init__(self, obj, params_hdl):
        """
        Paramters
        ---------
        obj : object
            model object

        params_hdl : UTransformersHandler
            Pipeline parameters handler during fit, fit_transform, transform function execution.
        """
        self.obj = obj
        self.params_hdl = params_hdl

    def __getattribute__(self, name):
        """
        This function return the attribute using the following the rules:
        - for `fit` function: return a wrapped function that update the input arguments with the cache and then call fit function.
        - for any other function we check:
            - if the estimator has the attribute -> invoke it
            - else -> try to invoke the function from the Wrapper class.
        """
        if name.startswith("fit"):

            def fitwrapper(X, y=None, **kargs):
                # kargs.update(getattr(object.__getattribute__(self, 'params_hdl'), 'get_estimator_paramters')())
                kargs.update(self.params_hdl.get_estimator_paramters())
                fit_params = {}
                if y is not None:
                    fit_params.update({"y": y})
                fit_params.update(kargs)
                return getattr(object.__getattribute__(self, "obj"), name)(
                    X, **fit_params
                )

            output = fitwrapper
        else:
            if hasattr(object.__getattribute__(self, "obj"), name):
                output = getattr(object.__getattribute__(self, "obj"), name)
            else:
                output = object.__getattribute__(self, name)
        return output


class EstimatorHandler:
    """
    This class handles actions realted to the estimator step (westimator).
    The class wraps the estimator and allows update the estimator parameters during a pipeline execution.
    """

    def __init__(self, params_hdl):
        self.params_hdl = params_hdl

    def wrap_and_link_estimator_step(self, steps):
        """
        Modify the estimator step and link step name and estimator.

        Description
        ----------
        This function wrap the estimator, save the step name and the westimator reference and
        return the updated steps list.

        Parameters
        ----------
        steps : list
            pipeline steps list

        Returns
        -------
        list

        """
        self.estimator_name = steps[-1][0]
        self.params_hdl.create_estimator_parameters(self.estimator_name, steps[-1][1])
        steps[-1] = (steps[-1][0], WEstimator(steps[-1][1], self.params_hdl))
        self.estimator = steps[-1][1]
        return steps

    def run_predictions(self, Xt):
        kargs = {}

        y_pred = self.estimator.predict(Xt)
        kargs["y_pred"] = y_pred

        if hasattr(self.estimator, "predict_proba"):
            y_proba = self.estimator.predict_proba(Xt)
            kargs["y_proba"] = y_proba
        return kargs



import numpy as np

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost
from holisticai.utils.transformers.bias import SensitiveGroups

from .randomized_threshold.algorithm import RandomizedThresholdAlgorithm
from .reduce2binary.algorithm import Reduce2BinaryAlgorithm


class MLDebiaser(BMPost):
    """
    MLDebiaser postprocessing debias predictions w.r.t. the sensitive class in
    each demographic group. This procedure takes as input a vector y and solves
    the optimization problem subject to the statistical parity constraint. This
    bias mitigation can be used for classification (binary and multiclass).

    Reference
    ---------
        Alabdulmohsin, Ibrahim M., and Mario Lucic. "A near-optimal algorithm for debiasing
        trained machine learning models." Advances in Neural Information Processing Systems
        34 (2021): 8072-8084.
    """

    def __init__(
        self,
        gamma=1.0,
        eps=0,
        eta=0.5,
        sgd_steps=10_000,
        full_gradient_epochs=1_000,
        batch_size=256,
        max_iter=5,
        verbose=True,
    ):
        self.gamma = gamma
        self.eps = eps
        self.eta = eta
        self.sgd_steps = sgd_steps
        self.full_gradient_epochs = full_gradient_epochs
        self.batch_size = batch_size
        self.max_iter = max_iter
        self.verbose = verbose
        self.sens_groups = SensitiveGroups()

    def fit(self, y_proba: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Compute parameters for calibrated equalized odds.

        Description
        ----------
        Compute parameters for calibrated equalized odds algorithm.

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (num_examples,).
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
        Self
        """
        params = self._load_data(y_proba=y_proba, group_a=group_a, group_b=group_b)
        y_proba = params["y_proba"]
        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        num_classes = y_proba.shape[1]
        sensitive_features = np.stack([group_a, group_b], axis=1)
        self.sens_groups.fit(sensitive_features)

        if num_classes > 2:
            self.algorithm = Reduce2BinaryAlgorithm(
                gamma=self.gamma,
                eps=self.eps,
                eta=self.eta,
                num_classes=num_classes,
                sgd_steps=self.sgd_steps,
                full_gradient_epochs=self.full_gradient_epochs,
                batch_size=self.batch_size,
                max_iter=self.max_iter,
                verbose=self.verbose,
            )
        else:
            self.algorithm = RandomizedThresholdAlgorithm(
                gamma=self.gamma,
                eps=self.eps,
                sgd_steps=self.sgd_steps,
                full_gradient_epochs=self.full_gradient_epochs,
                batch_size=self.batch_size,
                verbose=self.verbose,
            )
        return self

    def transform(
        self,
        y_proba: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Apply transform function to predictions and likelihoods

        Description
        ----------
        Use a fitted probability to change the output label and invert the likelihood

        Parameters
        ----------
        y_proba : array-like
            Predicted probability matrix (nb_examlpes, nb_classes)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        threshold : float
            float value to discriminate between 0 and 1

        Returns
        -------
        dictionnary with new predictions
        """
        params = self._load_data(y_proba=y_proba, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_proba = params["y_proba"]
        sensitive_features = np.stack([group_a, group_b], axis=1)
        p_attr = self.sens_groups.transform(sensitive_features, convert_numeric=True)

        if type(self.algorithm) is Reduce2BinaryAlgorithm:
            # Multiclass classification
            new_y_prob = self.algorithm.predict(y_proba, p_attr)
            new_y_pred = new_y_prob.argmax(axis=-1)
            return {"y_pred": new_y_pred, "y_proba": new_y_prob}
        else:
            # Binary classification
            pred = y_proba[:, 1]  # .argmax(axis=-1)
            pred = (
                2 * pred - 1
            )  # follow author implementation (use prediction and not logit)
            self.algorithm.fit(pred, p_attr)

            new_y_score = self.algorithm.predict(pred, p_attr)
            new_y_pred = np.where(new_y_score > 0.5, 1, 0)
            return {"y_pred": new_y_pred, "y_score": new_y_score}

    def fit_transform(
        self, y_proba: np.ndarray, group_a: np.ndarray, group_b: np.ndarray
    ):
        """
        Fit and transform

        Description
        ----------
        Fit and transform

        Parameters
        ----------
        y_proba : array-like
            Predicted vector (num_examples,).
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
        dictionnary with new predictions
        """
        return self.fit(
            y_proba,
            group_a,
            group_b,
        ).transform(y_proba, group_a, group_b)



from typing import Optional

import numpy as np
from sklearn.base import BaseEstimator

from holisticai.utils.transformers.bias import BMInprocessing as BMImp
from holisticai.utils.transformers.bias import SensitiveGroups

from .algorithm import KMediamClusteringAlgorithm


class FairKmedianClustering(BaseEstimator, BMImp):
    """
    Fair K-median clustering inprocessing bias mitigation is an approximation algorithms for
    group representative k-median clustering. Implementation of Algorithm 2. from [1]. The
    algorithm reduces the max group cost function.

    Reference:
    ----------
        [1] Abbasi, Mohsen, Aditya Bhaskara, and Suresh Venkatasubramanian. "Fair clustering via
        equitable group representations." Proceedings of the 2021 ACM Conference on Fairness,
        Accountability, and Transparency. 2021.
    """

    def __init__(
        self,
        n_clusters: int = 2,
        max_iter: int = 1000,
        seed: int = None,
        strategy: Optional[str] = "LS",
        verbose: Optional[int] = 0,
    ):
        """
        Parameters
        ----------

        n_clusters: int
            number of clusters.

        max_iter: int
            Max number of iteration for LS or epochs for GA.

        strategy: str
            Minimization method used. Available:
            - LS (Local Search).
            - GA (Genetic Algorithm)

        verbose: int
            if > 0 , print progress information.
        """
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.verbose = verbose
        self.seed = seed
        self.strategy = strategy
        self.sensgroup = SensitiveGroups()
        self.algorithm = KMediamClusteringAlgorithm(
            n_clusters=n_clusters, max_iter=max_iter, strategy=strategy, verbose=verbose
        )

    def fit(self, X, group_a, group_b):
        """
        Fit model using Fair K-median Clustering.

        Parameters
        ----------
        X : matrix-like
            Input matrix
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
            Self
        """
        np.random.seed(self.seed)

        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        sensitive_groups = np.c_[group_a, group_b]
        p_attr = np.array(
            self.sensgroup.fit_transform(sensitive_groups, convert_numeric=True)
        )

        self.algorithm.fit(X, p_attr)
        return self

    @property
    def labels_(self):
        return self.algorithm.labels_

    @property
    def cluster_centers_(self):
        return self.algorithm.cluster_centers_



import numpy as np
from scipy.optimize import fmin_cg

from holisticai.utils.transformers.bias import SensitiveGroups


class PrejudiceRemoverAlgorithm:
    """Two class LogisticRegression with Prejudice Remover"""

    def __init__(self, estimator, objective_fn, logger, maxiter=None):
        """
        Parameters
        ----------
        estimator : estimator object
            Prejudice Remover Model

        objective_fn : object
            Objective class with loss and grad_loss function

        logger : object
            Support for print information

        maxiter: int
            Maximum number of iterations for nonlinear conjugate gradient algorithm. Default is ``200 * len(x0)``.

        """

        self.estimator = estimator
        self.logger = logger
        self.objective_fn = objective_fn
        self.maxiter = maxiter
        self.sens_groups = SensitiveGroups()

    def fit(self, X: np.ndarray, y: np.ndarray, sensitive_features: np.ndarray):
        """
        Description
        -----------
        Optimize the model paramters to reduce the loss function

        Parameters
        ----------
        X : matrix-like
            Input matrix

        y_true : numpy array
            Target vector

        sensitive_features : numpy array
            Matrix where each columns is a sensitive feature e.g. [col_1=group_a, col_2=group_b]
        """

        groups_num = self.sens_groups.fit_transform(
            sensitive_features, convert_numeric=True
        )
        self.estimator.init_params(X, y, groups_num)
        self.logger.set_log_fn(
            loss=lambda coef: self.objective_fn.loss(coef, X, y, groups_num), type=float
        )

        self.coef = fmin_cg(
            self.objective_fn.loss,
            self.estimator.coef,
            fprime=self.objective_fn.grad_loss,
            args=(X, y, groups_num),
            maxiter=self.maxiter,
            disp=False,
            callback=self.logger.callback,
        )
        self.estimator.set_params(self.coef)

        self.f_loss_ = self.objective_fn.loss(self.coef, X, y, groups_num)
        self.logger.info(f"Best Loss : {self.f_loss_:.4f}")
        return self

    def predict(self, X, sensitive_features):
        """
        predict classes

        Parameters
        ----------
        X : array, shape=(n_samples, n_features)
            feature vectors of samples

        sensitive_features : numpy array
            Matrix where each columns is a sensitive feature e.g. [col_1=group_a, col_2=group_b]
        Returns
        -------
        y : array, shape=(n_samples), dtype=int
            array of predicted class
        """
        p_attr = self.sens_groups.transform(sensitive_features, convert_numeric=True)
        return self.estimator.predict(X, p_attr)

    def predict_proba(self, X, sensitive_features):
        """
        Predict probabilities

        Parameters
        ----------
        X : array, shape=(n_samples, n_features)
            feature vectors of samples

        sensitive_features : numpy array
            Matrix where each columns is a sensitive feature e.g. [col_1=group_a, col_2=group_b]

        Returns
        -------
        y_proba : array, shape=(n_samples, n_classes), dtype=float
            array of predicted class
        """
        p_attr = self.sens_groups.transform(sensitive_features, convert_numeric=True)
        return self.estimator.predict_proba(X, p_attr)



from typing import Optional

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin, clone

from holisticai.utils.transformers.bias import BMInprocessing as BMImp

from ..commons.classification import _constraints as cc
from ..commons.regression import _constraints as rc
from ..commons.regression import _losses as rl
from ._grid_generator import GridGenerator
from .algorithm import GridSearchAlgorithm


class GridSearchReduction(BaseEstimator, BMImp):
    """
    Grid search technique can be used for fair classification or fair regression.
    - For classification it reduces fair classification to a sequence of cost-sensitive classification problems,
    returning the deterministic classifier with the lowest empirical error subject to fair classification constraints among the
    candidates searched.
    - For regression it uses the same priniciple to return a deterministic regressor with the lowest empirical error subject to the
    constraint of bounded group loss.

    References:
        Agarwal, Alekh, et al. "A reductions approach to fair classification."
        International Conference on Machine Learning. PMLR, 2018.

        Agarwal, Alekh, Miroslav Dudík, and Zhiwei Steven Wu.
        "Fair regression: Quantitative definitions and reduction-based algorithms."
        International Conference on Machine Learning. PMLR, 2019.
    """

    CONSTRAINTS = [
        "DemographicParity",
        "EqualizedOdds",
        "TruePositiveRateParity",
        "FalsePositiveRateParity",
        "ErrorRateParity",
        "BoundedGroupLoss",
    ]

    def __init__(
        self,
        constraints: str = "EqualizedOdds",
        constraint_weight: Optional[float] = 0.5,
        loss: str = "ZeroOne",
        min_val: float = None,
        max_val: float = None,
        grid_size: Optional[int] = 10,
        grid_limit: Optional[float] = 2.0,
        verbose: Optional[int] = 0.0,
    ):
        """
        Init Grid Search Reduction Transformer

        Parameters
        ----------

        constraints : string
            The disparity constraints expressed as string:
                - "DemographicParity",
                - "EqualizedOdds",
                - "TruePositiveRateParity",
                - "FalsePositiveRateParity",
                - "ErrorRateParity"
                - "BoundedGroupLoss"

        constraint_weight : float
            Specifies the relative weight put on the constraint violation when selecting the
            best model. The weight placed on the error rate will be :code:`1-constraint_weight`

        loss : str
            String identifying loss function for constraints. Options include "ZeroOne", "Square", and "Absolute."

        min_val : float
            Loss function parameter for "Square" and "Absolute," typically the minimum of the range of y values.

        max_val: float
            Loss function parameter for "Square" and "Absolute," typically the maximum of the range of y values.

        grid_size : int
            The number of Lagrange multipliers to generate in the grid

        grid_limit : float
            The largest Lagrange multiplier to generate. The grid will contain
            values distributed between :code:`-grid_limit` and :code:`grid_limit`
            by default

        verbose : int
            If >0, will show progress percentage.

        """
        self.constraints = constraints
        self.constraint_weight = constraint_weight
        self.loss = loss
        self.min_val = min_val
        self.max_val = max_val
        self.grid_size = grid_size
        self.grid_limit = grid_limit
        self.verbose = verbose

    def transform_estimator(self, estimator):
        self.estimator = estimator
        return self

    def fit(
        self,
        X: np.ndarray,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """

        Fit model using Grid Search Reduction.

        Parameters
        ----------
        X : matrix-like
            Input matrix
        y_true : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
            Self
        """

        params = self._load_data(X=X, y_true=y_true, group_a=group_a, group_b=group_b)
        group_a = params["group_a"]
        group_b = params["group_b"]
        X = params["X"]
        y_true = params["y_true"]

        sensitive_features = np.stack([group_a, group_b], axis=1)

        self.estimator_ = clone(self.estimator)

        constraints_catalog = {
            "DemographicParity": cc.DemographicParity,
            "EqualizedOdds": cc.EqualizedOdds,
            "TruePositiveRateParity": cc.TruePositiveRateParity,
            "FalsePositiveRateParity": cc.FalsePositiveRateParity,
            "ErrorRateParity": cc.ErrorRateParity,
            "BoundedGroupLoss": rc.BoundedGroupLoss,
        }

        constraint_kargs = self._constraint_parameters()
        self.constraint_ = constraints_catalog[self.constraints](**constraint_kargs)

        self.model_ = GridSearchAlgorithm(
            estimator=self.estimator_,
            constraint=self.constraint_,
            constraint_weight=self.constraint_weight,
            grid_size=self.grid_size,
            grid_limit=self.grid_limit,
            verbose=self.verbose,
        )

        self.model_.fit(X, y_true, sensitive_features=sensitive_features)

        return self

    def _constraint_parameters(self):
        kargs = {}
        if self.constraints == "BoundedGroupLoss":
            losses = {
                "ZeroOne": rl.ZeroOneLoss,
                "Square": rl.SquareLoss,
                "Absolute": rl.AbsoluteLoss,
            }
            if self.loss == "ZeroOne":
                self.loss_ = losses[self.loss]()
            else:
                self.loss_ = losses[self.loss](self.min_val, self.max_val)
            kargs.update({"loss": self.loss_})
        return kargs

    def predict(self, X):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : matrix-like
            Input Matrix

        Returns
        -------
        numpy.ndarray: Predicted output
        """
        return self.model_.predict(X)

    def predict_proba(self, X):
        """
        Probability Prediction

        Description
        ----------
        Probability estimate for the given samples.

        Parameters
        ----------
        X : matrix-like
            Input Matrix

        Returns
        -------
        numpy.ndarray
            probability output
        """
        return self.model_.predict_proba(X)



import numpy as np

from holisticai.bias.mitigation.inprocessing.commons import Logging


class MFLogger:
    def __init__(self, total_iterations, verbose):
        self.verbose = verbose
        log_params = [
            ("iteration", int),
            ("fairness_error", float),
            ("fair_cluster_energy", float),
            ("cluster_energy", float),
        ]
        self.total_iterations = total_iterations
        self.logger = Logging(
            log_params=log_params,
            total_iterations=self.total_iterations,
            logger_format="iteration",
        )

    def update(self, iteration, fairness_error, fair_cluster_energy, cluster_energy):
        if self.verbose:
            self.logger.update(
                iteration, fairness_error, fair_cluster_energy, cluster_energy
            )



from typing import Union

import numpy as np

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost

from .algorithm import Algorithm


class MCMF(BMPost):
    """
    Minimal Cluster Modification for Fairnes (MCMF) is focused on the minimal change it so that the clustering is still
    of good quality and fairer.

    References:
        Davidson, Ian, and S. S. Ravi. "Making existing clusterings fairer: Algorithms, complexity results and insights."
        Proceedings of the AAAI Conference on Artificial Intelligence. Vol. 34. No. 04. 2020.
    """

    def __init__(
        self, metric: str = "L1", solver: str = "highs", group_mode="a", verbose=0
    ):
        """
        Create a Equalized Odds Post-processing instance.

        Parameters
        ----------
        metric : str
            Measure function used in the objective function.
            The metrics available are:
            ["constant", "L1", "L2"]

        solver : str
            Algorithm name used to solve the standard form problem. Solver supported must depend of your scipy poackage version.
            for scipy 1.9.0 the solvers available are:
            ["highs", "highs-ds", "highs-ipm"]

        group_mode: str
            Set what groups will be fitted: ['a', 'b', 'ab']
        verbose : int
            If > 0 , then print logs.
        """
        self.metric = metric
        self.group_mode = group_mode
        self.solver = solver
        self.verbose = verbose
        self.algorithm = Algorithm(metric=metric, solver=solver, verbose=verbose)

    def fit(self):
        return self

    def fit_transform(
        self,
        X: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        centroids: np.ndarray,
    ) -> "MCMF":
        return self.transform(X, y_pred, group_a, group_b, centroids)

    def transform(
        self,
        X: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        centroids: Union[np.ndarray, str],
    ) -> np.ndarray:
        """
        Modify y_pred using MCMF algorithm.

        Description
        ----------
        Build parameters for the objetive function and call the solver to find the algorithm parameters.

        Parameters
        ----------
        X : matrix-like
            Input matrix (nb_examlpes, nb_features)
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        Self
        """
        params = self._load_data(X=X, y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        X = params["X"]
        y_pred = params["y_pred"]

        if type(centroids) is str:
            centroids = getattr(self.estimator_hdl.estimator, centroids)

        if self.group_mode in ["a", "b"]:
            group = group_a if self.group_mode == "a" else group_b
            new_y_pred = self.algorithm.transform(
                X=X, y_pred=y_pred, group=group, centroids=centroids
            )
        elif self.group_mode == "ab":
            new_y_pred = y_pred.copy()
            for group in [group_a, group_b]:
                new_y_pred = self.algorithm.transform(
                    X=X, y_pred=new_y_pred, group=group, centroids=centroids
                )

        return {"y_pred": new_y_pred}



# Base Imports
import numpy as np
import pandas as pd
from sklearn import metrics

from ._validation import _array_like_to_numpy


def extract_columns(df, cols):
    """
    From pandas df to numpy arrays

    Description
    ----------
    This function returns numpy arrays for each column in cols

    Parameters
    ----------
    df : pandas DataFrame
        The data
    cols : list of str
        The column names

    Returns
    -------
    list of numpy arrays
    """
    out = []
    for c in cols:
        out.append(df[c].to_numpy())
    return out


def extract_group_vectors(p_attr, groups):
    """
    Extract Group Vectors

    Description
    ----------
    This function returns a list of group membership vectors
    (one for each group).

    Parameters
    ----------
    p_attr : array-like
        Protected attribute vector
    groups : list of str
        Group names

    Returns
    -------
    list of membership vectors
    """
    p_attr = _array_like_to_numpy(p_attr)
    out = []
    for g in groups:
        out.append(p_attr == g)
    return out


def mat_to_binary(mat, top=None, thresh=0.5):
    """
    Formatting helper function - makes a matrix binary

    Description
    ----------
    Given a recommender matrix, it will coerce to the binary form
    (shown / not shown). This will be first according to the top k rule
    if top is not None. If top is None, it will threshold scores at
    thresh (by default 0.5).

    Parameters
    ----------
    mat : numpy ndarray
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each item.
    top : int
        If not None, the top k scores that are shown to each user
    thresh : float
        Float in (0,1) range indicating threshold at which
        a given item is shown to user

    Returns
    -------
    float
        Binary matrix
    """
    # Case 1 : it's binary
    if np.array_equal(mat, mat.astype(bool)):
        return mat

    # Case 2 : top is not None
    if top is not None:
        # preprocess
        mat = np.nan_to_num(mat, copy=True, nan=-np.Inf)
        n_users, n_items = mat.shape
        top_items = np.argsort(mat, axis=1)

        # find indices of top items for each user
        columns = top_items[:, -top:].flatten()
        rows = np.arange(0, n_users).repeat(top, axis=0)

        # create
        n_mat = np.zeros((n_users, n_items))
        n_mat[rows, columns] = 1

        # impute nans with 0
        n_mat[np.isnan(n_mat)] = 0

    # Case 3 : the score is thresholded
    else:
        n_mat = mat >= thresh

    return n_mat * 1


def normalize_tensor(tensor):
    """
    Formatting helper function - normalises a tensor to [0,1] range

    Description
    ----------
    Given a numpy ndarray, we return a normalised array so that all
    values are in [0,1] range

    Parameters
    ----------
    tensor : numpy ndarray
        a numpy ndarray

    Returns
    -------
    numpy ndarray
        Normalised tensor
    """
    a = np.nanmin(tensor)
    b = np.nanmax(tensor)
    return (tensor - a) / (b - a)


def slice_arrays_by_quantile(q, scores, arr_ls):
    """
    Slice arrays by quantile

    Description
    ----------
    Slices to above given quantile for list of arrays.

    Parameters
    ----------
    q : numpy array
        Quantile
    scores : numpy array
        The scores to slice according to the quantile q
    arr_ls :
        The list of arrays to slice

    Returns
    -------
    list of numpy arrays
        sliced_arrays
    """

    # Get the top indices according to quantile
    scores = np.array(scores)
    top = np.quantile(scores, q)
    top_ind = scores.reshape(1, -1) >= top.reshape(-1, 1)

    sliced_arrays = [[arr[i] for i in top_ind] for arr in arr_ls]

    return sliced_arrays


def recommender_formatter(
    df, users_col, groups_col, items_col, scores_col, aggfunc="sum"
):
    """
    Recommender formatter

    Description
    ----------
    Coerce pandas (users-groups-items-scores) dataframe to format we use
    for recommender bias metrics: pivoted matrix of scores with shape
    (num_users, num_items).

    Parameters
    ----------
    df : pandas DataFrame
        Dataframe with recommender data
    users_col : str
        The name of the colum with user id's
    groups_col : str
        The name of the column with group information
    items_col : str
        The name of the column with item id's
    scores_col : str
        The name of the column with score
        (predicted or true) of user on given item
    aggfunc : 'sum' or 'mean'
        the aggregation function for duplicate index, column pairs

    Returns
    -------
    list of numpy ndarray
        df_pivot, p_attr
    """
    # pivot dataframe on users and items
    df_pivot = df.pivot_table(
        index=users_col, columns=items_col, values=scores_col, aggfunc=aggfunc
    )
    # we need to get group info for each user
    user_to_group = dict(zip(df[users_col], df[groups_col]))
    p_attr = np.array(df_pivot.index.map(user_to_group))

    return df_pivot, p_attr



# Base Imports
import numpy as np
import pandas as pd

# utils
from ...utils._formatting import slice_arrays_by_quantile
from ...utils._validation import _check_non_empty, _regression_checks


def _calc_success_rate(group_membership: np.array, threshold=float):
    return (group_membership > threshold).mean()


def success_rate_regression(group_a, group_b, y_pred, threshold=0.50):
    """
    Calculates the raw success rates for each group.
    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (numerical)
    threshold: float, str
        The number above which the result is considered a success. Also accepts 'median' and 'mean'.
    Returns
    -------
    dict
        Dictionary with two keys, sr_a and sr_b (success rate for group a and b)
    """
    # Needs to be numpy array or the following operations won't be correct
    if (type(threshold) == str) and (threshold not in set(["median", "mean"])):
        raise ValueError("Threshold not recognised")
    if threshold == "median":
        threshold = np.median(y_pred)
    if threshold == "mean":
        threshold = np.mean(y_pred)
    group_a = np.array(group_a)
    group_b = np.array(group_b)
    y_pred = np.array(y_pred)
    group_a_membership = y_pred[group_a == 1]
    group_b_membership = y_pred[group_b == 1]
    sr_a = _calc_success_rate(group_a_membership, threshold)
    sr_b = _calc_success_rate(group_b_membership, threshold)  # success rate group_b
    return {"sr_a": sr_a, "sr_b": sr_b}


def disparate_impact_regression(group_a, group_b, y_pred, q=0.8):
    r"""
    Disparate Impact quantile (Regression version).

    Description
    -----------
    This function computes the ratio of success rates between group_a and
    group_b, where sucess means predicted score exceeds a given quantile (default = 0.8).

    If q is a vector, this function returns a vector with the
    respective result for each given quantile in q.

    Interpretation
    --------------
    A value of 1 is desired. Values below 1 are unfair towards group_a.
    Values above 1 are unfair towards group_b. The range (0.8,1.2)
    is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    q (optional) : float or array, default=0.8
        quantile of predictions considered

    Returns
    -------
    float
        Disparate Impact (top %) : :math:`\frac{sr_a}{sr_b}`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import disparate_impact_regression
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([0.8, 0.9, 0.2, 0.1, 0.7, 0.9, 0.8, 0.6, 0.3, 0.5])
    >>> disparate_impact_regression(group_a, group_b, y_pred, q=0.7)
    1.5
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, q = _regression_checks(
        group_a, group_b, y_pred, None, q
    )

    # numbers
    n_a = group_a.sum()
    n_b = group_b.sum()

    _, group_a_list, group_b_list = slice_arrays_by_quantile(
        q, y_pred, [y_pred, group_a, group_b]
    )

    n_q = len(group_a_list)
    disp_impact = np.zeros(n_q)

    for i in range(n_q):
        group_a, group_b = group_a_list[i], group_b_list[i]
        _check_non_empty(group_a, name="group_a", quantile=q[i])
        _check_non_empty(group_b, name="group_b", quantile=q[i])

        sr_a = group_a.sum() / n_a  # success rate group_a
        sr_b = group_b.sum() / n_b  # success rate group_b

        disp_impact[i] = sr_a / sr_b

    return np.squeeze(disp_impact)[()]


def statistical_parity_regression(group_a, group_b, y_pred, q=0.5):
    """
    Statistical Parity quantile (Regression version).

    Description
    -----------
    This function computes the difference of success rates between group_a and
    group_b, where sucess means that the predicted score exceeds a given quantile.

    If q is a vector, this function returns a vector with the
    respective result for each given quantile in q.

    Interpretation
    --------------
    A value of 0 is desired. Values below 0 are unfair towards group_a.
    Values above 0 are unfair towards group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    q (optional) : float or array, default=0.5
        quantile of predictions considered

    Returns
    -------
    float
        Statistical Parity (top %) : SR_a - SR_b

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import statistical_parity_regression
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([0.8, 0.9, 0.2, 0.1, 0.7, 0.9, 0.8, 0.6, 0.3, 0.5])
    >>> statistical_parity_regression(group_a, group_b, y_pred, q=0.7)
    0.16666666666666669
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, q = _regression_checks(
        group_a, group_b, y_pred, None, q
    )

    # numbers
    n_a = group_a.sum()
    n_b = group_b.sum()

    _, group_a_list, group_b_list = slice_arrays_by_quantile(
        q, y_pred, [y_pred, group_a, group_b]
    )

    n_q = len(group_a_list)
    stat_parity = np.zeros(n_q)

    for i in range(n_q):
        group_a, group_b = group_a_list[i], group_b_list[i]
        _check_non_empty(group_a, name="group_a", quantile=q[i])
        _check_non_empty(group_b, name="group_b", quantile=q[i])

        sr_a = group_a.sum() / n_a  # success rate group_a
        sr_b = group_b.sum() / n_b  # success rate group_b

        stat_parity[i] = sr_a - sr_b

    return np.squeeze(stat_parity)[()]


def no_disparate_impact_level(group_a, group_b, y_pred):
    """
    No disparate impact level.

    Description
    -----------
    This function computes the maximum score such that thresholding at that score
    does not allow adverse impact.

    Interpretation
    --------------

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)

    Returns
    -------
    float
        No disparate impact level

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import no_disparate_impact_level
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([0.8, 0.9, 0.2, 0.1, 0.7, 0.9, 0.8, 0.6, 0.3, 0.5])
    >>> no_disparate_impact_level(group_a, group_b, y_pred)
    0.7
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _ = _regression_checks(
        group_a, group_b, y_pred, None, None
    )

    # grid
    q = np.linspace(1.0, 0.0, 100)

    # try different values
    for v in q:
        pred = np.quantile(y_pred, v)
        pass_members = y_pred >= pred
        a = sum(group_a * pass_members) / group_a.sum()
        b = sum(group_b * pass_members) / group_b.sum()
        # find score that does not allow adverse impact
        if b > 0:
            if 0.8 < a / b < 1.2:
                break
    return pred


def avg_score_diff(group_a, group_b, y_pred, q=0):
    """

    Average Score Difference.

    Description
    -----------
    This function computes the difference in average scores between
    group_a and group_b.

    If q is a vector, this function returns a vector with the
    respective result for each given quantile in q.

    Interpretation
    --------------
    A value of 0 is desired. Negative values indicate the group_a
    has lower average score, so bias against group_a. Positive values
    indicate group_b has lower average score, so bias against group_b.
    Scale is relative to task.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    q (optional) : float or array
        quantile of predictions considered

    Returns
    -------
    float
        Average Score Spread : :math:`\texttt{AVgroup_a - AVgroup_b}`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import avg_score_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([0.8, 0.9, 0.2, 0.1, 0.7, 0.9, 0.8, 0.6, 0.3, 0.5])
    >>> avg_score_diff(group_a, group_b, y_pred)
    -0.13333333333333341
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, q = _regression_checks(
        group_a, group_b, y_pred, None, q
    )

    y_pred_list, group_a_list, group_b_list = slice_arrays_by_quantile(
        q, y_pred, [y_pred, group_a, group_b]
    )

    n_q = len(group_a_list)
    avg_score_diff = np.zeros(n_q)

    for i in range(n_q):
        group_a, group_b = group_a_list[i], group_b_list[i]
        _check_non_empty(group_a, name="group_a", quantile=q[i])
        _check_non_empty(group_b, name="group_b", quantile=q[i])

        y_pred = y_pred_list[i]
        avgroup_a = y_pred[group_a == 1].mean()
        avgroup_b = y_pred[group_b == 1].mean()

        avg_score_diff[i] = avgroup_a - avgroup_b

    return np.squeeze(avg_score_diff)[()]


def avg_score_ratio(group_a, group_b, y_pred, q=0):
    """

    Average Score Ratio.

    Description
    -----------
    This function computes the ratio in average scores between
    group_a and group_b.

    If q is a vector, this function returns a vector with the
    respective result for each given quantile in q.

    Interpretation
    --------------
    A value of 1 is desired. Values below 1 indicate the group_a
    has lower average score, so bias against group_a. Values above 1
    indicate group_b has lower average score, so bias against group_b.
    (0.8, 1.25) range is considered fair.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    q (optional) : float or array
        quantile of predictions considered

    Returns
    -------
    float
        Average Score Ratio : :math:`\texttt{AVgroup_a / AVgroup_b}`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import avg_score_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([0.8, 0.9, 0.2, 0.1, 0.7, 0.9, 0.8, 0.6, 0.3, 0.5])
    >>> avg_score_ratio(group_a, group_b, y_pred)
    0.7894736842
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, q = _regression_checks(
        group_a, group_b, y_pred, None, q
    )

    y_pred_list, group_a_list, group_b_list = slice_arrays_by_quantile(
        q, y_pred, [y_pred, group_a, group_b]
    )

    n_q = len(group_a_list)
    avg_score_ratio = np.zeros(n_q)

    for i in range(n_q):
        group_a, group_b = group_a_list[i], group_b_list[i]
        _check_non_empty(group_a, name="group_a", quantile=q[i])
        _check_non_empty(group_b, name="group_b", quantile=q[i])

        y_pred = y_pred_list[i]
        avgroup_a = y_pred[group_a == 1].mean()
        avgroup_b = y_pred[group_b == 1].mean()

        avg_score_ratio[i] = avgroup_a / avgroup_b

    return np.squeeze(avg_score_ratio)[()]


def zscore_diff(group_a, group_b, y_pred, q=0):
    """

    ZScore Difference.

    Description
    -----------
    This function computes the spread in Zscores between
    group_a and group_b. The Zscore is a normalised
    version of Disparate Impact.

    If q is a vector, this function returns a vector with the
    respective result for each given quantile in q.

    Interpretation
    --------------
    A value of 0 is desired. The Zscore will approximate the number
    of standard deviations away from the mean. In particular values that
    exceed 2 are statistically significant with 95% probability.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    q (optional) : float or array
        quantile of predictions considered

    Returns
    -------
    float
        ZScore Difference : :math:`\frac{\texttt{AVgroup_a} - \texttt{AVgroup_b}}{\texttt{STD_pool}}`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import zscore_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([0.8, 0.9, 0.2, 0.1, 0.2, 0.9, 0.3, 0.6, 0.3, 0.5])
    >>> zscore_diff(group_a, group_b, y_pred)
    0.1166919931983158
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, q = _regression_checks(
        group_a, group_b, y_pred, None, q
    )

    y_pred_list, group_a_list, group_b_list = slice_arrays_by_quantile(
        q, y_pred, [y_pred, group_a, group_b]
    )

    n_q = len(group_a_list)
    zscore_diff = np.zeros(n_q)

    for i in range(n_q):
        group_a, group_b = group_a_list[i], group_b_list[i]
        _check_non_empty(group_a, name="group_a", quantile=q[i])
        _check_non_empty(group_b, name="group_b", quantile=q[i])

        # Get means
        y_pred = y_pred_list[i]
        av_group_a = y_pred[group_a == 1].mean()
        av_group_b = y_pred[group_b == 1].mean()

        # Get n_a and n_b
        n_a = group_a.sum()
        n_b = group_b.sum()

        # STD pooled
        std_pool = np.sqrt(
            (
                (n_a - 1) * y_pred[group_a == 1].std() ** 2.0
                + (n_b - 1) * y_pred[group_b == 1].std() ** 2.0
            )
            / (n_b + n_a - 2)
        )

        zscore_diff[i] = (av_group_a - av_group_b) / std_pool

    return np.squeeze(zscore_diff)[()]


def statistical_parity_auc(group_a, group_b, y_pred):
    """
    Statistical parity (AUC).

    Description
    -----------
    This function computes the area under the statistical parity
    versus threshold curve.

    Interpretation
    --------------
    A value of 0 is desired. Values below 0.075 are considered
    acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)

    Returns
    -------
    float
        statistical parity (AUC)

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import statistical_parity_auc
    >>> group_a = np.array([1] * 50 + [0] * 50)
    >>> group_b = np.array([0] * 50 + [1] * 50)
    >>> y_pred = np.concatenate((np.linspace(-1, 1, 50), np.linspace(-1, 1, 50)**3))
    >>> statistical_parity_auc(group_a, group_b, y_pred)
    0.12106666666666668
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _ = _regression_checks(
        group_a, group_b, y_pred, None, None
    )

    # thresholds
    thresh = np.linspace(1, 0, 150)
    pass_value = np.quantile(y_pred, thresh)
    y_binary = y_pred.reshape(-1, 1) >= pass_value.reshape(1, -1)
    pass_a = y_binary[group_a == 1].sum(axis=0) / group_a.sum()
    pass_b = y_binary[group_b == 1].sum(axis=0) / group_b.sum()
    di_arr = np.abs(pass_a - pass_b)

    # AUC
    return np.sum(di_arr * np.array([1 / 150] * 150))


def _weighed_statistical_parity_auc(group_a, group_b, y_pred):
    """
    Weighed Statistical parity (AUC).

    Description
    -----------
    This function computes the area under the statistical
    parity versus threshold curve, weighed by the 2t distribution.

    Interpretation
    --------------
    A value of 0 is desired. Values below 0.1 are considered
    acceptable.

    Parameters
    ----------
    group_a : numpy array
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)

    Returns
    -------
    float
        Weighed statistical parity (AUC)

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import _weighed_statistical_parity_auc
    >>> group_a = np.array([1] * 50 + [0] * 50)
    >>> group_b = np.array([0] * 50 + [1] * 50)
    >>> y_pred = np.concatenate((np.linspace(-1, 1, 50), np.linspace(-1, 1, 50)**3))
    >>> _weighed_statistical_parity_auc(group_a, group_b, y_pred)
    0.12106666666666666
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _ = _regression_checks(
        group_a, group_b, y_pred, None, None
    )

    # thresholds
    thresh = np.linspace(1, 0, 150)
    pass_value = np.quantile(y_pred, thresh)
    y_binary = y_pred.reshape(-1, 1) >= pass_value.reshape(1, -1)
    pass_a = y_binary[group_a == 1].sum(axis=0) / group_a.sum()
    pass_b = y_binary[group_b == 1].sum(axis=0) / group_b.sum()
    di_arr = np.abs(pass_a - pass_b)
    differentials = np.linspace(2, 0, 150)

    # Weighed AUC
    return np.sum(di_arr * differentials / 150)


def max_statistical_parity(group_a, group_b, y_pred):
    """
    Max absolute statistical parity.

    Description
    -----------
    This function computes the maximum over all thresholds of
    the absolute statistical parity between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. Values below 0.1 in absolute value are
    considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)

    Returns
    -------
    float
        max absolute statistical parity

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import max_statistical_parity
    >>> group_a = np.array([1] * 50 + [0] * 50)
    >>> group_b = np.array([0] * 50 + [1] * 50)
    >>> y_pred = np.concatenate((np.linspace(-1, 1, 50), np.linspace(-1, 1, 50)**3))
    >>> max_statistical_parity(group_a, group_b, y_pred)
    0.20000000000000007
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _ = _regression_checks(
        group_a, group_b, y_pred, None, None
    )

    # thresholds
    thresh = np.linspace(1, 0, 150)
    pass_value = np.quantile(y_pred, thresh)
    y_binary = y_pred.reshape(-1, 1) >= pass_value.reshape(1, -1)
    pass_a = y_binary[group_a == 1].sum(axis=0) / group_a.sum()
    pass_b = y_binary[group_b == 1].sum(axis=0) / group_b.sum()
    di_arr = np.abs(pass_a - pass_b)

    # MAX
    return np.max(di_arr)


def correlation_diff(group_a, group_b, y_pred, y_true, q=0):
    """
    Correlation difference.

    Description
    -----------
    This function computes the difference in correlation between predictions
    and targets for group_a and group_b.

    If q is a vector, this function returns a vector with the
    respective result for each given quantile in q.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -2 and 2,
    with -1 indicating strong bias against group_a, and +1
    indicating strong bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    y_true : numpy array
        Target vector (regression)
    q (optional) : float or array
        quantile of predictions considered

    Returns
    -------
    float
        correlation difference : :math:`CV_a - CV_b`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import correlation_diff
    >>> group_a = np.array([1] * 50 + [0] * 50)
    >>> group_b = np.array([0] * 50 + [1] * 50)
    >>> y_pred = np.concatenate((np.linspace(-1, 1, 50), np.linspace(-1, 1, 50)**3))
    >>> y_true = np.concatenate((np.linspace(-1, 1, 50), np.linspace(-1, 1, 50)**2))
    >>> correlation_diff(group_a, group_b, y_pred, y_true, q=0)
    1.0000000000000002
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, q = _regression_checks(
        group_a, group_b, y_pred, None, q
    )

    y_pred_list, y_true_list, group_a_list, group_b_list = slice_arrays_by_quantile(
        q, y_pred, [y_pred, y_true, group_a, group_b]
    )

    n_q = len(group_a_list)
    corr_diff = np.zeros(n_q)

    for i in range(n_q):
        group_a, group_b = group_a_list[i], group_b_list[i]
        _check_non_empty(group_a, name="group_a", quantile=q[i])
        _check_non_empty(group_b, name="group_b", quantile=q[i])

        # Compute Pearson correlations
        y_pred = y_pred_list[i]
        y_true = y_true_list[i]
        cv_a = np.corrcoef(y_pred[group_a == 1], y_true[group_a == 1])[1, 0]
        cv_b = np.corrcoef(y_pred[group_b == 1], y_true[group_b == 1])[1, 0]

        corr_diff[i] = cv_a - cv_b

    return np.squeeze(corr_diff)[()]


def rmse_ratio(group_a, group_b, y_pred, y_true, q=0):
    """
    RMSE ratio.

    Description
    -----------
    This function computes the ratio of the RMSE for group_a and group_b.

    If q is a vector, this function returns a vector with the
    respective result for each given quantile in q.

    Interpretation
    --------------
    A value of 1 is desired. Lower values show bias against group_a.
    Higher values show bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    y_true : numpy array
        Target vector (regression)
    q (optional) : float or array
        quantile of predictions considered

    Returns
    -------
    float
        RMSE ratio : :math:`\frac{RMSE_a}{RMSE_b}`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import rmse_ratio
    >>> group_a = np.array([1] * 50 + [0] * 50)
    >>> group_b = np.array([0] * 50 + [1] * 50)
    >>> y_pred = np.concatenate((np.linspace(-1, 1, 50), np.linspace(-1, 1, 50)))
    >>> y_true = np.concatenate((np.linspace(-1, 1, 50)**2, np.linspace(-1, 1, 50)**3))
    >>> rmse_ratio(group_a, group_b, y_pred, y_true)
    2.7471209467641367
    """
    # check and coerce inputs
    group_a, group_b, y_pred, y_true, q = _regression_checks(
        group_a, group_b, y_pred, y_true, q
    )

    y_pred_list, y_true_list, group_a_list, group_b_list = slice_arrays_by_quantile(
        q, y_pred, [y_pred, y_true, group_a, group_b]
    )

    n_q = len(group_a_list)
    rmse_ratio = np.zeros(n_q)

    for i in range(n_q):
        group_a, group_b = group_a_list[i], group_b_list[i]
        _check_non_empty(group_a, name="group_a", quantile=q[i])
        _check_non_empty(group_b, name="group_b", quantile=q[i])

        # Compute RMSE for both groups
        y_pred = y_pred_list[i]
        y_true = y_true_list[i]
        rmse_a = np.sqrt(((y_true[group_a == 1] - y_pred[group_a == 1]) ** 2.0).mean())
        rmse_b = np.sqrt(((y_true[group_b == 1] - y_pred[group_b == 1]) ** 2.0).mean())

        rmse_ratio[i] = rmse_a / rmse_b

    return np.squeeze(rmse_ratio)[()]


def mae_ratio(group_a, group_b, y_pred, y_true, q=0):
    """
    MAE ratio

    Description
    -----------
    This function computes the ratio of the MAE for group_a and group_b.

    If q is a vector, this function returns a vector with the
    respective result for each given quantile in q.

    Interpretation
    --------------
    A value of 1 is desired. Lower values show bias against group_a.
    Higher values show bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    y_true : numpy array
        Target vector (regression)
    q (optional) : float or array
        quantile of predictions considered

    Returns
    -------
    float
        MAE ratio : :math:`\frac{MAE_a}{MAE_b}`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import mae_ratio
    >>> group_a = np.array([1] * 50 + [0] * 50)
    >>> group_b = np.array([0] * 50 + [1] * 50)
    >>> y_pred = np.concatenate((np.linspace(-1, 1, 50), np.linspace(-1, 1, 50)))
    >>> y_true = np.concatenate((np.linspace(-1, 1, 50)**2, np.linspace(-1, 1, 50)**3))
    >>> mae_ratio(group_a, group_b, y_pred, y_true)
    2.084201388888889
    """
    # check and coerce inputs
    group_a, group_b, y_pred, y_true, q = _regression_checks(
        group_a, group_b, y_pred, y_true, q
    )

    y_pred_list, y_true_list, group_a_list, group_b_list = slice_arrays_by_quantile(
        q, y_pred, [y_pred, y_true, group_a, group_b]
    )

    n_q = len(group_a_list)
    mae_ratio = np.zeros(n_q)

    for i in range(n_q):
        group_a, group_b = group_a_list[i], group_b_list[i]
        _check_non_empty(group_a, name="group_a", quantile=q[i])
        _check_non_empty(group_b, name="group_b", quantile=q[i])

        # Compute MAE for both groups
        y_pred = y_pred_list[i]
        y_true = y_true_list[i]
        mae_a = (np.abs(y_true[group_a == 1] - y_pred[group_a == 1])).mean()
        mae_b = (np.abs(y_true[group_b == 1] - y_pred[group_b == 1])).mean()

        mae_ratio[i] = mae_a / mae_b

    return np.squeeze(mae_ratio)[()]


def regression_bias_metrics(
    group_a, group_b, y_pred, y_true=None, metric_type="equal_outcome"
):
    """
    Regression bias metrics batch computation

    Description
    -----------
    This function computes all the relevant regression bias metrics,
    and displays them as a pandas dataframe.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (regression)
    y_true (optional) : numpy array
        Target vector (regression)
    metric_type : 'both', 'equal_outcome' or 'equal_opportunity'
        Specifies which metrics we compute

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """

    equal_outcome_metrics = {
        "Disparate Impact Q90": disparate_impact_regression,
        "Disparate Impact Q80": disparate_impact_regression,
        "Disparate Impact Q50": disparate_impact_regression,
        "Statistical Parity Q50": statistical_parity_regression,
        "No Disparate Impact Level": no_disparate_impact_level,
        "Average Score Difference": avg_score_diff,
        "Average Score Ratio": avg_score_ratio,
        "Z Score Difference": zscore_diff,
        "Max Statistical Parity": max_statistical_parity,
        "Statistical Parity AUC": statistical_parity_auc,
    }

    equal_opportunity_metrics = {
        "RMSE Ratio": rmse_ratio,
        "RMSE Ratio Q80": rmse_ratio,
        "MAE Ratio": mae_ratio,
        "MAE Ratio Q80": mae_ratio,
        "Correlation Difference": correlation_diff,
    }

    ref_vals = {
        "Disparate Impact Q90": 1,
        "Disparate Impact Q80": 1,
        "Disparate Impact Q50": 1,
        "Statistical Parity Q50": 0,
        "No Disparate Impact Level": "-",
        "Average Score Difference": 0,
        "Average Score Ratio": 1,
        "Z Score Difference": 0,
        "Max Statistical Parity": 0,
        "Statistical Parity AUC": 0,
        "RMSE Ratio": 1,
        "RMSE Ratio Q80": 1,
        "MAE Ratio": 1,
        "MAE Ratio Q80": 1,
        "Correlation Difference": 0,
    }

    hypers = {
        "Disparate Impact Q90": {"q": 0.9},
        "Disparate Impact Q80": {"q": 0.8},
        "Disparate Impact Q50": {"q": 0.5},
        "Statistical Parity Q50": {"q": 0.8},
        "No Disparate Impact Level": {},
        "Average Score Difference": {},
        "Average Score Ratio": {},
        "Z Score Difference": {},
        "Max Statistical Parity": {},
        "Statistical Parity AUC": {},
        "RMSE Ratio": {},
        "RMSE Ratio Q80": {"q": 0.8},
        "MAE Ratio": {},
        "MAE Ratio Q80": {"q": 0.8},
        "Correlation Difference": {},
    }

    out_metrics = [
        [pf, fn(group_a, group_b, y_pred, **hypers[pf]), ref_vals[pf]]
        for pf, fn in equal_outcome_metrics.items()
    ]
    if y_true is not None:
        opp_metrics = [
            [pf, fn(group_a, group_b, y_pred, y_true, **hypers[pf]), ref_vals[pf]]
            for pf, fn in equal_opportunity_metrics.items()
        ]

    if metric_type == "both":
        metrics = out_metrics + opp_metrics
        return pd.DataFrame(
            metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    elif metric_type == "equal_outcome":
        return pd.DataFrame(
            out_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    elif metric_type == "equal_opportunity":
        return pd.DataFrame(
            opp_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    else:
        raise ValueError(
            "metric_type is not one of : both, equal_outcome, equal_opportunity"
        )



import numpy as np
import pandas as pd

from ._conventions import _EVENT, _GROUP_ID, _LABEL, _SIGNED


def merge_columns(feature_columns):
    return pd.DataFrame(feature_columns).apply(
        lambda row: ",".join([str(r) for r in row.values]), axis=1
    )


def format_data(y=None):
    new_y = pd.Series(np.array(y).reshape(-1))
    return {"y": new_y}


class BaseMoment:
    """
    Base Moment

    Abstract class used to used to describe:
    - Optimization Objective
    - Fairness Constraints
    """

    params = ["X", "y", "sensitive_features", "control_features"]

    def save_params(self, *args):
        for name, value in zip(self.params, args):
            setattr(self, name, value)

    def load_data(self, X, y, sensitive_features):
        self.tags = pd.DataFrame()
        self.tags[_LABEL] = y
        self.tags[_GROUP_ID] = merge_columns(sensitive_features)
        self.save_params(X, y, sensitive_features)

    @property
    def total_samples(self):
        """Return the number of samples in the data."""
        return self.X.shape[0]

    @property
    def _y_as_series(self):
        """Return the y array as a :class:`~pandas.Series`."""
        return self.y

    def project_lambda(self, lambda_vec):
        """
        Projected lambda values.

        Returns
        -------
        Returns lambda which is guaranteed to lead to the same or higher value of the
        Lagrangian compared with lambda_vec for all possible choices of the classifier, h.
        """
        if self.ratio == 1.0:
            lambda_pos = lambda_vec["+"] - lambda_vec["-"]
            lambda_neg = -lambda_pos
            lambda_pos[lambda_pos < 0.0] = 0.0
            lambda_neg[lambda_neg < 0.0] = 0.0
            lambda_projected = pd.concat(
                [lambda_pos, lambda_neg],
                keys=["+", "-"],
                names=[_SIGNED, _EVENT, _GROUP_ID],
            )
            return lambda_projected
        return lambda_vec

    def bound(self):
        """
        Return bound vector.

        Returns
        -------
        pandas.Series
            a vector of bound values corresponding to all constraints

        """
        return pd.Series(self.eps, index=self.index)



import math

import numpy as np
from joblib import Parallel, delayed
from numpy.linalg import inv


def calculate_popularity_model(ratings):
    propensity_score = []
    for i in range(ratings.shape[1]):
        propensity_score.append(
            float(np.count_nonzero(ratings[:, i])) / ratings.shape[0]
        )

    temp = np.array(propensity_score)
    temp = temp.reshape((1, len(temp)))
    return temp


def erros(P_s, Q_s, R_s, W_s, lamda, beta):
    eR_s = np.sum(P_s * Q_s.T, axis=1)
    et = np.sum(pow(R_s - eR_s, 2))
    e_s = (lamda / 2) * (pow(P_s, 2) + pow(Q_s.T, 2)) + (beta / 2) * (
        pow((pow(P_s, 2) + pow(Q_s.T, 2)), 2) * W_s[:, None]
    )
    countt = len(e_s)
    e = et + np.sum(e_s)
    err = math.sqrt(e / countt)
    errt = math.sqrt(et / countt)
    return err, errt


def compute_mat(qr, ivp):
    qr_ = qr[:, :, None]
    qr_t = qr[:, None, :]
    ivp_ = ivp.reshape(-1, 1, 1)
    return np.sum(ivp_ * np.matmul(qr_, qr_t), axis=0)


def compute_vect(qr, ivp, r):
    temp = (ivp * r).reshape(-1, 1)
    return np.sum(temp * qr, axis=0)


def update_emb(b, R_, E_, invprop_):
    mat_temp = compute_mat(E_, invprop_)
    vect_temp = compute_vect(E_, invprop_, R_)
    return np.dot(inv(mat_temp + b), vect_temp)


def updateP(P, b, R, Q, u_rated_items, invprop):
    P = Parallel(n_jobs=-1, verbose=0)(
        delayed(update_emb)(
            b, R[u, rated_items], Q[:, rated_items].T, invprop[u, rated_items]
        )
        for u, rated_items in enumerate(u_rated_items)
    )
    return np.stack(P, axis=0)


def updateQ(P, b, R, Q, i_rated_users, invprop):
    Q = Parallel(n_jobs=-1, verbose=0)(
        delayed(update_emb)(
            b, R[rated_users, i], P[rated_users, :], invprop[rated_users, i]
        )
        for i, rated_users in enumerate(i_rated_users)
    )
    return np.stack(Q, axis=1)



from time import time

import numpy as np
import pandas as pd
import scipy.optimize as opt
from sklearn import clone

from ..commons._conventions import PRECISION


class Lagrangian:
    """Operations related to the Lagrangian"""

    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        estimator,
        constraints,
        B: float,
        opt_lambda: bool = True,
        **kwargs
    ):
        """
        Initialize Lagrangian object.

        Parameters
        ----------
        X :  matrix-like
            the training features
        sensitive_features :  array-like
            the sensitive features to use for constraints
        y :  array-like
            the training labels
        estimator :
            the estimator to fit in every iteration of :meth:`best_h` using a
            :meth:`fit` method with arguments `X`, `y`, and `sample_weight`
        constraints : ClassificationConstraint
            Object describing the parity constraints. This provides the reweighting
            and relabelling.
        B : float
            bound on the L1-norm of the lambda vector
        opt_lambda : bool
            indicates whether to optimize lambda during the calculation of the
            Lagrangian; optional with default value True
        """
        self.constraints = constraints
        self.constraints.load_data(X, y, **kwargs)
        self.obj = self.constraints.default_objective()
        self.obj.load_data(X, y, **kwargs)
        self.estimator = estimator
        self.B = B
        self.opt_lambda = opt_lambda
        self.hs = pd.Series(dtype="float64")
        self.predictors = pd.Series(dtype="float64")
        self.errors = pd.Series(dtype="float64")
        self.gammas = pd.DataFrame()
        self.lambdas = pd.DataFrame()
        self.last_linprog_n_hs = 0
        self.last_linprog_result = None

    def _eval(self, Q, lambda_vec):
        """
        Return the value of the Lagrangian.

        Parameters
        ----------
        Q : {pandas.Series, callable}
            `Q` is either a series of weights summing up to 1 that indicate
            the weight of each `h` in contributing to the randomized
            predictor, or a callable corresponding to a deterministic
            `predict` function.
        lambda_vec : pandas.Series
            lambda vector

        Returns
        -------
        tuple
            tuple `(L, L_high, gamma, error)` where `L` is the value of the
            Lagrangian, `L_high` is the value of the Lagrangian under the best
            response of the lambda player, `gamma` is the vector of constraint
            violations, and `error` is the empirical error
        """
        if callable(Q):
            error = self.obj.gamma(Q)[0]
            gamma = self.constraints.gamma(Q)
        else:
            error = self.errors[Q.index].dot(Q)
            gamma = self.gammas[Q.index].dot(Q)

        if self.opt_lambda:
            lambda_projected = self.constraints.project_lambda(lambda_vec)
            L = error + np.sum(lambda_projected * (gamma - self.constraints.bound()))
        else:
            L = error + np.sum(lambda_vec * (gamma - self.constraints.bound()))

        max_constraint = (gamma - self.constraints.bound()).max()
        if max_constraint <= 0:
            L_high = error
        else:
            L_high = error + self.B * max_constraint
        return L, L_high, gamma, error

    def eval_gap(self, Q, lambda_hat, nu):
        r"""Return the duality gap object for the given :math:`Q` and :math:`\hat{\lambda}`."""
        L, L_high, gamma, error = self._eval(Q, lambda_hat)
        result = _GapResult(L, L, L_high, gamma, error)
        for mul in [1.0, 2.0, 5.0, 10.0]:
            h_hat, h_hat_idx = self.best_h(mul * lambda_hat)
            L_low_mul, _, _, _ = self._eval(pd.Series({h_hat_idx: 1.0}), lambda_hat)
            if L_low_mul < result.L_low:
                result.L_low = L_low_mul
            if result.gap() > nu + PRECISION:
                break
        return result

    def solve_linprog(self, nu):
        n_hs = len(self.hs)
        n_constraints = len(self.constraints.index)
        if self.last_linprog_n_hs == n_hs:
            return self.last_linprog_result
        c = np.concatenate((self.errors, [self.B]))
        A_ub = np.concatenate(
            (
                self.gammas.sub(self.constraints.bound(), axis=0),
                -np.ones((n_constraints, 1)),
            ),
            axis=1,
        )
        b_ub = np.zeros(n_constraints)
        A_eq = np.concatenate((np.ones((1, n_hs)), np.zeros((1, 1))), axis=1)
        b_eq = np.ones(1)
        result = opt.linprog(
            c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, method="simplex"
        )
        Q = pd.Series(result.x[:-1], self.hs.index)
        dual_c = np.concatenate((b_ub, -b_eq))
        dual_A_ub = np.concatenate((-A_ub.transpose(), A_eq.transpose()), axis=1)
        dual_b_ub = c
        dual_bounds = [
            (None, None) if i == n_constraints else (0, None)
            for i in range(n_constraints + 1)
        ]  # noqa: E501
        result_dual = opt.linprog(
            dual_c, A_ub=dual_A_ub, b_ub=dual_b_ub, bounds=dual_bounds, method="simplex"
        )
        lambda_vec = pd.Series(result_dual.x[:-1], self.constraints.index)
        self.last_linprog_n_hs = n_hs
        self.last_linprog_result = (Q, lambda_vec, self.eval_gap(Q, lambda_vec, nu))
        return self.last_linprog_result

    def _call_oracle(self, lambda_vec):
        signed_weights = self.obj.signed_weights() + self.constraints.signed_weights(
            lambda_vec
        )
        if self.constraints.PROBLEM_TYPE == "classification":
            y = 1 * (signed_weights > 0)
        else:
            y = self.constraints._y_as_series

        w = signed_weights.abs()
        sample_weight = self.constraints.total_samples * w / w.sum()

        estimator = clone(estimator=self.estimator, safe=False)
        estimator.fit(self.constraints.X, y, **{"sample_weight": sample_weight})
        return estimator

    def best_h(self, lambda_vec):
        """
        Solve the best-response problem.

        Description
        -----------
        Returns the classifier that solves the best-response problem for
        the vector of Lagrange multipliers `lambda_vec`.
        """
        classifier = self._call_oracle(lambda_vec)

        def h(X):
            pred = classifier.predict(X)
            # Some estimators return an output of the shape (num_preds, 1) - flatten such
            # results
            if getattr(pred, "flatten", None) is not None:
                pred = pred.flatten()
            return pred

        h_error = self.obj.gamma(h)[0]
        h_gamma = self.constraints.gamma(h)
        h_value = h_error + h_gamma.dot(lambda_vec)

        if not self.hs.empty:
            values = self.errors + self.gammas.transpose().dot(lambda_vec)
            best_idx = values.idxmin()
            best_value = values[best_idx]
        else:
            best_idx = -1
            best_value = np.PINF

        if h_value < best_value - PRECISION:
            h_idx = len(self.hs)
            self.hs.at[h_idx] = h
            self.predictors.at[h_idx] = classifier
            self.errors.at[h_idx] = h_error
            self.gammas[h_idx] = h_gamma
            self.lambdas[h_idx] = lambda_vec.copy()
            best_idx = h_idx

        return self.hs[best_idx], best_idx


class _GapResult:
    """The result of a duality gap computation."""

    def __init__(self, L, L_low, L_high, gamma, error):
        self.L = L
        self.L_low = L_low
        self.L_high = L_high
        self.gamma = gamma
        self.error = error

    def gap(self):
        return max(self.L - self.L_low, self.L_high - self.L)
